"""
A module of useful math for handling coordinate transformations and things
"""
import itertools
import math

import numpy as np
from . import Misc as util
from .Options import Options

__all__ = [
    "vec_dots",
    "vec_handle_zero_norms",
    "vec_apply_zero_threshold",
    "vec_normalize",
    "vec_rescale",
    "vec_norms",
    "vec_tensordot",
    "vec_tensordiag",
    "vec_block_diag",
    "diag_indices",
    "block_array",
    "vec_tdot",
    "distance_matrix",
    "identity_tensors",
    "semisparse_tensordot",
    "frac_powh",
    "vec_crosses",
    "vec_angles",
    "vec_sins",
    "vec_cos",
    "vec_outer",
    "pts_norms",
    "pts_angles",
    "pts_normals",
    "vec_dihedrals",
    "pts_dihedrals",
    "pts_book_angles",
    "mat_vec_muls",
    "one_pad_vecs",
    "affine_multiply",
    "cartesian_from_rad",
    "polar_to_cartesian",
    "apply_by_coordinates",
    "apply_by_structures",
    "find_basis",
    "projection_matrix",
    "orthogonal_projection_matrix",
    "project_onto",
    "project_out",
    "fractional_power",
    "unitarize_transformation",
    "maximum_similarity_transformation",
    "polar_decomposition",
    "matrix_transform_from_eigs",
    "symmetric_matrix_exp",
    "imaginary_symmetric_matrix_exp",
    "symmetric_matrix_log",
    "imaginary_symmetric_matrix_log",
    "sylvester_solve",
    "symmetrize_array"
]

##
# TODO: The design of a lot of this stuff needs a bit of work
#       Like should it work with things that aren't just stacks of vectors?
#       Or should it all be specifically for vector-vector operations?
#       Lots of it doesn't even make sense in a non-vector context...
#       But then there's also like "vec_tensordot" which is explicitly non-vector in scope...
#       Not sure what exactly I want with this. Lots of stuff TBD.

################################################
#
#       vec_dots
#
def vec_dots(vecs1, vecs2, axis=-1):
    """
    Computes the pair-wise dot product of two lists of vecs using np.matmul

    :param vecs1:
    :type vecs1:
    :param vecs2:
    :type vecs2:
    """

    vecs1 = np.expand_dims(vecs1, axis-1)
    vecs2 = np.expand_dims(vecs2, axis)
    res = np.matmul(vecs1, vecs2)
    res_shape = res.shape

    for _ in range(2):
        if res_shape[axis] == 1:
            res = res.reshape(np.delete(res_shape, axis))
            res_shape = res.shape

    return res

################################################
#
#       vec_norms
#
def vec_norms(vecs, axis=-1):
    """

    :param vecs:
    :type vecs: np.ndarray
    :param axis:
    :type axis: int
    :return:
    :rtype:
    """
    # if axis != -1:
    #     raise NotImplementedError("Norm along not-the-last axis not there yet...")
    return np.linalg.norm(vecs, axis=axis)

def distance_matrix(pts, axis=-1, axis2=None, return_triu=False, return_indices=False):
    pts = np.asanyarray(pts)
    if axis2 is None:
        axis2 = axis-1

    n = pts.shape[axis2]
    rows, cols = np.triu_indices(n, k=1)
    vecs_r = np.take(pts, rows, axis=axis2)
    vecs_c = np.take(pts, cols, axis=axis2)

    dists = vec_norms(vecs_r - vecs_c, axis=axis)
    if return_triu:
        if return_indices:
            return dists, (rows, cols)
        else:
            return dists
    else:
        dist_mats = np.zeros(tuple(np.delete(pts.shape, [axis, axis2])) + (n, n), dtype=dists.dtype)
        dist_mats[..., rows, cols] = dists
        dist_mats[..., cols, rows] = dists
        #TODO: handle transposition
        return dist_mats

################################################
#
#       vec_normalize
#
def vec_apply_zero_threshold(vecs, zero_thresh=None, return_zeros=False):
    """
    Applies a threshold to cast nearly-zero vectors to proper zero

    :param vecs:
    :type vecs:
    :param zero_thresh:
    :type zero_thresh:
    :return:
    :rtype:
    """
    norms = vec_norms(vecs)
    vecs, zeros = vec_handle_zero_norms(vecs, norms, zero_thresh=zero_thresh)
    norms = norms[..., np.newaxis]
    norms[zeros] = Options.zero_placeholder

    if return_zeros:
        return vecs, norms, zeros
    else:
        return vecs, norms

def vec_handle_zero_norms(vecs, norms, axis=-1, zero_thresh=None):
    """
    Tries to handle zero-threshold application to vectors

    :param vecs:
    :type vecs:
    :param norms:
    :type norms:
    :param zero_thesh:
    :type zero_thesh:
    :return:
    :rtype:
    """
    norms = np.expand_dims(norms, axis)
    zero_thresh = Options.norm_zero_threshold if zero_thresh is None else zero_thresh
    zeros = np.abs(norms) < zero_thresh
    vecs = vecs * (1 - zeros.astype(int))
    return vecs, zeros

def vec_normalize(vecs, norms=None, axis=-1, zero_thresh=None, return_norms=False):
    """

    :param vecs:
    :type vecs: np.ndarray
    :param axis:
    :type axis: int
    :return:
    :rtype:
    """
    if norms is None:
        norms = vec_norms(vecs, axis=axis)
    vecs, zeros = vec_handle_zero_norms(vecs, norms, axis=axis, zero_thresh=zero_thresh)
    e_norms = np.expand_dims(norms, axis)
    e_norms[zeros] = Options.zero_placeholder # since we already zeroed out the vector

    if return_norms:
        return vecs/e_norms, norms
    else:
        return vecs/e_norms

def vec_rescale(vecs, target_range=None, cur_range=None, axis=-1):
    vecs = np.asanyarray(vecs)
    if cur_range is None:
        cur_mins = np.expand_dims(np.min(vecs, axis=axis), axis)
        cur_max = np.expand_dims(np.max(vecs, axis=axis), axis)
        cur_range = cur_max - cur_mins
    else:
        cur_mins = cur_range[0]
        cur_range = cur_range[1] - cur_range[0]
    rescaled = (vecs - cur_mins) / cur_range
    if target_range is not None:
        if util.is_numeric(target_range):
            rescaled = rescaled * target_range
        else:
            t_min, t_max = target_range
            if not util.is_numeric(t_min):
                t_min = np.expand_dims(np.asanyarray(t_min), axis)
            if not util.is_numeric(t_max):
                t_max = np.expand_dims(np.asanyarray(t_max), axis)
            rescaled = rescaled*(t_max - t_min) + t_min
    return rescaled

################################################
#
#       vec_crosses
#

def vec_crosses(vecs1, vecs2, normalize=False, zero_thresh=None, axis=-1):
    crosses = np.cross(vecs1, vecs2, axis=axis)
    if normalize:
        norms = vec_norms(crosses, axis=axis)
        zero_thresh = Options.norm_zero_threshold if zero_thresh is None else zero_thresh
        smol = util.is_numeric(norms)
        if smol:
            bad_norms = norms < zero_thresh
            if bad_norms:
                norms = np.array(1.)
        else:
            bad_norms = np.where(norms <= zero_thresh)
            norms[bad_norms] = 1.

        crosses = crosses/norms[..., np.newaxis]

        if smol:
            if bad_norms:
                crosses *= 0.
        else:
            crosses[bad_norms] *= 0.

    return crosses

################################################
#
#       vec_cos
#
def vec_cos(vectors1, vectors2, zero_thresh=None, axis=-1):
    """Gets the cos of the angle between two vectors

    :param vectors1:
    :type vectors1: np.ndarray
    :param vectors2:
    :type vectors2: np.ndarray
    """
    dots   = vec_dots(vectors1, vectors2, axis=axis)
    norms1 = vec_norms(vectors1, axis=axis)
    norms2 = vec_norms(vectors2, axis=axis)

    norm_prod = norms1 * norms2
    if isinstance(norm_prod, np.ndarray):
        zero_thresh = Options.norm_zero_threshold if zero_thresh is None else zero_thresh
        bad_norm_prods = np.where(np.abs(norm_prod) <= zero_thresh)
        norm_prod[bad_norm_prods] = 1.

    coses = dots/(norms1*norms2)

    if isinstance(norm_prod, np.ndarray):
        coses[bad_norm_prods] = 0.

    return coses

################################################
#
#       vec_sins
#
def vec_sins(vectors1, vectors2, zero_thresh=None, axis=-1):
    """Gets the sin of the angle between two vectors

    :param vectors1:
    :type vectors1: np.ndarray
    :param vectors2:
    :type vectors2: np.ndarray
    """
    crosses= vec_crosses(vectors1, vectors2, axis=axis)
    norms1 = vec_norms(vectors1, axis=axis)
    norms2 = vec_norms(vectors2, axis=axis)

    norm_prod = norms1 * norms2
    if isinstance(norm_prod, np.ndarray):
        zero_thresh = Options.norm_zero_threshold if zero_thresh is None else zero_thresh
        bad_norm_prods = np.where(np.abs(norm_prod) <= zero_thresh)
        norm_prod[bad_norm_prods] = 1.

    sines = vec_norms(crosses)/norm_prod

    if isinstance(norm_prod, np.ndarray):
        sines[bad_norm_prods] = 0.

    return sines


################################################
#
#       vec_angles
#
def vec_angles(vectors1, vectors2, norms=None, up_vectors=None, zero_thresh=None, axis=-1,
               return_norms=False,
               return_crosses=True,
               check_zeros=True
               ):
    """
    Gets the angles and normals between two vectors

    :param vectors1:
    :type vectors1: np.ndarray
    :param vectors2:
    :type vectors2: np.ndarray
    :param up_vectors: orientation vectors to obtain signed angles
    :type up_vectors: None | np.ndarray
    :return: angles and normals between two vectors
    :rtype: (np.ndarray, np.ndarray)
    """
    dots = vec_dots(vectors1, vectors2, axis=axis)
    crosses = vec_crosses(vectors1, vectors2, axis=axis)
    if norms is not None:
        norms1, norms2 = norms
    else:
        norms1 = vec_norms(vectors1, axis=axis)
        norms2 = vec_norms(vectors2, axis=axis)
        norms = (norms1, norms2)

    norm_prod = norms1*norms2
    if check_zeros:
        zero_thresh = Options.norm_zero_threshold if zero_thresh is None else zero_thresh
        smol = util.is_numeric(norm_prod)
        if smol:
            bad_norms = norm_prod < zero_thresh
            if bad_norms:
                norm_prod = np.array(1.)
        else:
            bad_norm_prods = np.where(np.abs(norm_prod) <= zero_thresh)
            norm_prod[bad_norm_prods] = 1.

    cos_comps = dots/norm_prod
    cross_norms = vec_norms(crosses, axis=axis)
    sin_comps = cross_norms/norm_prod

    angles = np.arctan2(sin_comps, cos_comps)

    if check_zeros:
        if smol:
            if bad_norms:
                angles = np.array(0.)
        else:
            angles[bad_norm_prods] = 0.

    # return signed angles
    if up_vectors is not None:
        if up_vectors.ndim < crosses.ndim:
            up_vectors = np.broadcast_to(up_vectors, crosses.shape[:-len(up_vectors.shape)] + up_vectors.shape)
        orientations = np.sign(vec_dots(up_vectors, crosses))
        angles = orientations * angles

    if return_crosses or return_norms:
        ret = (angles,)
        if return_crosses:
            ret = ret + (crosses,)
        if return_norms:
            ret = ret + (norms,)
    else:
        ret = angles

    return ret

################################################
#
#       vec_outer
#
def _vec_outer(a, b, axes=None):
    """
    Provides the outer product of a and b in a vectorized way.
    Currently not entirely convinced I'm doing it right :|

    :param a:
    :type a:
    :param b:
    :type b:
    :param axis:
    :type axis:
    :return:
    :rtype:
    """
    # we'll treat this like tensor_dot:
    #   first we turn this into a plain matrix
    #   then we do the outer on the matrix
    #   then we cast back to the shape we want
    if axes is None:
        if a.ndim > 1:
            axes = [-1, -1]
        else:
            axes = [0, 0]

    # we figure out how we'd conver
    a_ax = axes[0]
    if isinstance(a_ax, (int, np.integer)):
        a_ax = [a_ax]
    a_ax = [ax + a.ndim if ax<0 else ax for ax in a_ax]
    a_leftover = [x for x in range(a.ndim) if x not in a_ax]
    a_transp = a_leftover + a_ax
    a_shape = a.shape
    a_old_shape = [a_shape[x] for x in a_leftover]
    a_subshape = [a_shape[x] for x in a_ax]
    a_contract = a_old_shape + [np.prod(a_subshape)]

    b_ax = axes[1]
    if isinstance(b_ax, (int, np.integer)):
        b_ax = [b_ax]
    b_ax = [ax + b.ndim if ax<0 else ax for ax in b_ax]
    b_leftover = [x for x in range(b.ndim) if x not in b_ax]
    b_transp = b_leftover + b_ax
    b_shape = b.shape
    b_old_shape = [b_shape[x] for x in b_leftover]
    b_subshape = [b_shape[x] for x in b_ax]
    b_contract = b_old_shape + [np.prod(b_subshape)]

    a_new = a.transpose(a_transp).reshape(a_contract)
    b_new = b.transpose(b_transp).reshape(b_contract)

    if b_new.ndim < a_new.ndim:
        ...
    elif a_new.ndim < b_new.ndim:
        ...

    outer = a_new[..., :, np.newaxis] * b_new[..., np.newaxis, :]

    # now we put the shapes right again and revert the transposition
    # base assumption is that a_old_shape == b_old_shape
    # if not we'll get an error anyway
    final_shape = a_old_shape + a_subshape + b_subshape

    res = outer.reshape(final_shape)
    final_transp = np.argsort(a_leftover + a_ax + b_ax)

    return res.transpose(final_transp)

def bump_axes(a_ax, b_ax):
    new_b = []
    for _ in b_ax:  # pad b to get block form (a first, then b)
        while _ in a_ax or _ in new_b:
            _ = _ + 1
        new_b.append(_)
    return new_b
def riffle_axes(a_ax, b_ax):
    new_b = []
    new_a = []
    a_ax = list(a_ax)
    b_ax = list(b_ax)
    nmax = max(len(b_ax), len(a_ax))
    for n in range(nmax):
        if n < len(b_ax):
            bx = b_ax[n]
            if bx in a_ax:  # hit a collision, need to bump bx, then force a_ax to check for collisions next
                bx = bx + 1
                b_ax[n] = bx
                for m in range(n + 1, len(b_ax)):
                    if b_ax[m] == b_ax[m - 1]:
                        b_ax[m] += 1
                try:
                    bx_ax_pos = a_ax.index(bx)
                except:
                    pass
                else:
                    a_ax[bx_ax_pos] += 1
                    for m in range(bx_ax_pos + 1, len(a_ax)):
                        if a_ax[m] == a_ax[m - 1]:
                            a_ax[m] += 1
            new_b.append(bx)
        if n < len(a_ax):
            new_a.append(a_ax[n])
    return new_a, new_b
def vec_outer(a, b, axes=None, order=2):
    """
    Provides the outer product of a and b in a vectorized way.
    We have to prioritize what goes where, and `order` determines
    if the axes of `a` come first or `b` come first

    :param a:
    :type a:
    :param b:
    :type b:
    :param axis:
    :type axis:
    :return:
    :rtype:
    """
    if axes is None:
        if a.ndim > 1:
            axes = [-1, -1]
        else:
            axes = [0, 0]

    a_ax, b_ax = axes
    if isinstance(a_ax, (int, np.integer)): a_ax = [a_ax]
    if isinstance(b_ax, (int, np.integer)): b_ax = [b_ax]
    a_ax = [a.ndim + x if x < 0 else x for x in a_ax]
    b_ax = [b.ndim + x if x < 0 else x for x in b_ax]

    a_ax = list(np.sort(a_ax))
    b_ax = list(np.sort(b_ax))
    # we resolve axis conflicts, where two axes are mapping to the same spot
    if order == 0: # pad b to get block form (a first, then b)
        b_ax = bump_axes(a_ax, b_ax)
    elif order == 1: # pad a to get block form (b first, then a)
        a_ax = bump_axes(b_ax, a_ax)
    elif order == 2: # pad to get interleaving (a first, then b)
        a_ax, b_ax = riffle_axes(a_ax, b_ax)
    elif order == 3: # pad to get interleaving (a first, then b)
        a_ax, b_ax = riffle_axes(b_ax, a_ax)

    # we expand A and B appropriately now
    a = np.expand_dims(a, b_ax)
    b = np.expand_dims(b, a_ax)

    # and just multipy and let numpy broadcasting do the rest
    return a * b

################################################
#
#       vec_outer
#
def diag_indices(block_shape, n, k=2):
    ranges = [np.arange(s, dtype=int) for s in block_shape]
    indexing_shape = []
    m = len(block_shape)
    for i,r in enumerate(ranges):
        r = np.expand_dims(r, list(range(i)))
        r = np.expand_dims(r, [-x for x in range(1, m-i+1)])
        indexing_shape.append(r)
    di = np.expand_dims(np.arange(n, dtype=int), list(range(len(block_shape))))
    return tuple(indexing_shape) + (di,) * k
def vec_tensordiag(obj, axis=-1, extra_dims=1):
    obj = np.asanyarray(obj)
    base_shape = obj.shape[:axis]
    shp = obj.shape[axis:]
    extra_shp = (shp[0],) * extra_dims
    tensor = np.zeros(base_shape + extra_shp + shp, dtype=obj.dtype)
    inds = diag_indices(base_shape, shp[0], k=(extra_dims + 1))
    tensor[inds] = obj
    return tensor
def block_array(blocks, ndim=2, padding=0):
    blocks = np.asanyarray(blocks)
    for k in range(blocks.ndim-ndim-padding):
        blocks = np.concatenate(blocks, axis=-(k+1))
    return blocks
def vec_block_diag(mats):
    mats = np.asanyarray(mats)
    base_shape = mats.shape[:-3]
    stack_shape = mats.shape[-3]
    arr_shape = mats.shape[-2:]
    arr = np.zeros(base_shape + (stack_shape, stack_shape) + arr_shape)
    inds = diag_indices(base_shape, stack_shape, k=2)
    arr[inds] = mats
    return np.moveaxis(arr, -2, -3).reshape(base_shape + (stack_shape*arr_shape[-1], stack_shape*arr_shape[-2]))


def identity_tensors(base_shape, ndim):
    eye = np.eye(ndim)
    if util.is_int(base_shape):
        base_shape = [base_shape]
    base_shape = tuple(base_shape)
    return np.broadcast_to(
        np.expand_dims(eye, list(range(len(base_shape)))),
        base_shape + (ndim, ndim)
    )

#################################################################################
#
#   vec_tensordot
#
def vec_tensordot(tensa, tensb, axes=2, shared=None):
    """Defines a version of tensordot that uses matmul to operate over stacks of things
    Basically had to duplicate the code for regular tensordot but then change the final call

    :param tensa:
    :type tensa:
    :param tensb:
    :type tensb:
    :param axes:
    :type axes: [list[int]|int, list[int]|int] | int
    :param shared: the axes that should be treated as shared (for now just an int)
    :type shared: int | None
    :return:
    :rtype:
    """

    if isinstance(axes, (int, np.integer)):
        axes = (list(range(-axes, 0)), list(range(0, axes)))
    axes_a, axes_b = axes
    try:
        na = len(axes_a)
        axes_a = list(axes_a)
    except TypeError:
        axes_a = [axes_a]
        na = 1
    try:
        nb = len(axes_b)
        axes_b = list(axes_b)
    except TypeError:
        axes_b = [axes_b]
        nb = 1

    a, b = np.asarray(tensa), np.asarray(tensb)

    axes_a = [ax if ax >= 0 else a.ndim + ax for ax in axes_a]
    axes_b = [ax if ax >= 0 else b.ndim + ax for ax in axes_b]
    a_shape = tensa.shape
    b_shape = tensb.shape

    if shared is None:
        shared = 0
        for shared, s in enumerate(zip(a_shape, b_shape)):
            if s[0] != s[1]:
                break
            shared = shared + 1
    # else:

    # the minimum number of possible shared axes
    # is constrained by the contraction of axes
    shared = min([shared, min(axes_a), min(axes_b)])

    if shared == 0: #easier to just delegate here than handle more special cases
        return np.tensordot(a, b, axes=axes)

    as_ = a_shape
    nda = a.ndim
    bs = b.shape
    ndb = b.ndim

    equal = True
    if na != nb:
        equal = False
        raise ValueError("{}: shape-mismatch ({}) and ({}) in number of axes to contract over".format(
            "vec_tensordot",
            na,
            nb
        ))
    else:
        for k in range(na):
            if as_[axes_a[k]] != bs[axes_b[k]]:
                equal = False
                raise ValueError("{}: shape-mismatch ({}) and ({}) in contraction over axes ({}) and ({})".format(
                    "vec_tensordot",
                    axes_a[k],
                    axes_b[k],
                    na,
                    nb
                    ))
            if axes_a[k] < 0:
                axes_a[k] += nda
            if axes_b[k] < 0:
                axes_b[k] += ndb

    # Move the axes to sum over to the end of "a"
    # and to the front of "b"
    # preserve things so that the "shared" stuff remains at the fron of both of these...
    notin_a = [k for k in range(shared, nda) if k not in axes_a]
    newaxes_a = list(range(shared)) + notin_a + axes_a
    N2_a = 1
    for axis in axes_a:
        N2_a *= as_[axis]
    newshape_a = as_[:shared] + (int(np.prod([as_[ax] for ax in notin_a if ax >= shared])), N2_a)
    olda = [as_[axis] for axis in notin_a if axis >= shared]

    notin_b = [k for k in range(shared, ndb) if k not in axes_b]
    newaxes_b = list(range(shared)) + axes_b + notin_b
    N2_b = 1
    for axis in axes_b:
        N2_b *= bs[axis]
    newshape_b = as_[:shared] + (N2_b, int(np.prod([bs[ax] for ax in notin_b if ax >= shared])))
    oldb = [bs[axis] for axis in notin_b if axis >= shared]

    at = a.transpose(newaxes_a).reshape(newshape_a)
    bt = b.transpose(newaxes_b).reshape(newshape_b)
    res = np.matmul(at, bt)
    final_shape = list(a_shape[:shared]) + olda + oldb
    # raise Exception(res.shape, final_shape)
    return res.reshape(final_shape)
def vec_tdot(tensa, tensb, axes=((-1,), (1,))):
    """
    Tensor dot but just along the final axes by default. Totally a convenience function.

    :param tensa:
    :type tensa:
    :param tensb:
    :type tensb:
    :param axes:
    :type axes:
    :return:
    :rtype:
    """

    return vec_tensordot(tensa, tensb, axes=axes)

def semisparse_tensordot(sparse_data, a, /, axes, shared=None):
    pos, vals, shape = sparse_data
    a = np.asanyarray(a)
    lv_axes = axes[0]
    a_axes = axes[1]
    if util.is_numeric(lv_axes):
        lv_axes = [lv_axes]
    if util.is_numeric(a_axes):
        a_axes = [a_axes]


    if shared is None:
        shared = 0
    nax = len(lv_axes)
    for ax in reversed(a_axes):
        a = np.moveaxis(a, ax, shared)

    idx = np.setdiff1d(np.arange(len(shape)), lv_axes)

    target_shape = a.shape[:shared] + tuple(shape[i] for i in idx) + a.shape[nax + shared:]

    pos = [pos[i] for i in idx] + [pos[i] for i in lv_axes]
    shape = [shape[i] for i in idx] + [shape[i] for i in lv_axes]
    flat_shape = (np.prod([shape[i] for i in idx], dtype=int), np.prod([shape[i] for i in lv_axes], dtype=int))
    flat_pos = np.array(np.unravel_index(np.ravel_multi_index(pos, shape), flat_shape)).T

    a = np.reshape(a, a.shape[:shared] + (np.prod([a.shape[i+shared] for i in range(nax)], dtype=int), -1))

    # now we can do a form of sparse dot product

    ndim = len(shape) - shared
    if nax == ndim:
        idx = (slice(None),)*shared + (flat_pos,)
        new = vec_tensordot(
            vals,
            a[idx],
            axes=[0, shared]
        )
    else:
        new_shape = a.shape[:shared] + (flat_shape[0],) + a.shape[shared+1:]
        new = np.zeros(new_shape)
        pad_pos = (slice(None),)*shared
        for (i,k),v in zip(flat_pos, vals):
            i = pad_pos + (i,)
            k = pad_pos + (k,)
            new[i] += v * a[k]

    return np.reshape(new, target_shape)


def frac_powh(A, k, eigsys=None, pow=None):
    if eigsys is None:
        eigsys = np.linalg.eigh(A)
    eigvals, Q = eigsys
    if pow is None:
        pow = np.pow
    eigvals = pow(eigvals, k)
    # TODO: speed up with semisparse dot products
    if eigvals.ndim == 1:
        return Q @ np.diag(eigvals) @ Q.T
    else:
        shared = eigvals.ndim - 1
        v = vec_tensordiag(eigvals, extra_dims=shared)
        return vec_tensordot(vec_tensordot(Q, v, axes=[-1, -2], shared=shared), Q, axes=[-1, -2], shared=shared)
################################################
#
#       pts_norms
#
def pts_norms(pts1, pts2, **opts):
    """Provides the distance between the points

    :param pts1:
    :type pts1: np.ndarray
    :param pts2:
    :type pts2: np.ndarray
    :return:
    :rtype: np.ndarray
    """
    return vec_norms(pts2-pts1, **opts)

################################################
#
#       pts_angles
#
def pts_angles(pts1, pts2, pts3, **opts):
    """Provides the vector normal to the plane of the three points

    :param pts1:
    :type pts1: np.ndarray
    :param pts2:
    :type pts2: np.ndarray
    :param pts3:
    :type pts3: np.ndarray
    :return:
    :rtype: np.ndarray
    """
    return vec_angles(pts1-pts2, pts3-pts2, **opts)

################################################
#
#       pts_normals
#
def pts_normals(pts1, pts2, pts3, normalize=True, **opts):
    """Provides the vector normal to the plane of the three points

    :param pts1:
    :type pts1: np.ndarray
    :param pts2:
    :type pts2: np.ndarray
    :param pts3:
    :type pts3: np.ndarray
    :param normalize:
    :type normalize:
    :return:
    :rtype: np.ndarray
    """
    # should I normalize these...?
    return vec_crosses(pts2-pts1, pts3-pts1, normalize=normalize, **opts)

################################################
#
#       vec_dihedrals
#
def vec_dihedrals(b1, b2, b3,
                  crosses=None,
                  norms=None,
                  return_crosses=False
                  ):
    """
    Provides the dihedral angle between pts4 and the plane of the other three vectors

    :param pts1:
    :type pts1: np.ndarray
    :param pts2:
    :type pts2: np.ndarray
    :param pts3:
    :type pts3: np.ndarray
    :return:
    :rtype:
    """

    # # should I normalize these...?
    # normals = pts_normals(pts2, pts3, pts4, normalize=False)
    # off_plane_vecs = pts1 - pts4
    # return vec_angles(normals, off_plane_vecs)[0]

    # compute signed angle between the normals to the b1xb2 plane and b2xb3 plane
    # b1 = pts2-pts1 # 4->1
    # b2 = pts3-pts2 # 1->2
    # b3 = pts4-pts3 # 2->3

    if crosses is not None:
        crosses, cross_norms = crosses
        n1, n2 = crosses
        n1 = vec_normalize(n1, cross_norms[0])
        n2 = vec_normalize(n2, cross_norms[1])
    else:
        n1 = vec_crosses(b1, b2)#, normalize=True)
        n2 = vec_crosses(b2, b3)#, normalize=True)
        norm1 = vec_norms(n1)
        norm2 = vec_norms(n2)
        n1 = vec_normalize(n1, norm1)
        n2 = vec_normalize(n2, norm2)
        crosses = [(n1, n2), (norm1, norm2)]

    if norms is not None:
        nb1, nb2, nb3 = norms
        u2 = vec_normalize(b2, nb2)
    else:
        u2 = vec_normalize(b2)

    m1 = vec_crosses(n1, u2)
    d1 = vec_dots(n1, n2)
    d2 = vec_dots(m1, n2)

    # arctan(d2/d1) + sign stuff from relative signs of d2 and d1
    ret = -np.arctan2(d2, d1)
    if return_crosses:
        ret = (ret, crosses)
    return ret

################################################
#
#       pts_dihedrals
#
def pts_dihedrals(pts1, pts2, pts3, pts4,
                  crosses=None,
                  norms=None,
                  return_crosses=False, **opts
                  ):
    """
    Provides the dihedral angle between pts4 and the plane of the other three vectors

    :param pts1:
    :type pts1: np.ndarray
    :param pts2:
    :type pts2: np.ndarray
    :param pts3:
    :type pts3: np.ndarray
    :return:
    :rtype:
    """

    # b1 = pts2-pts1 # 4->1
    # b2 = pts3-pts2 # 1->2
    # b3 = pts4-pts3 # 2->3

    # # should I normalize these...?
    # normals = pts_normals(pts2, pts3, pts4, normalize=False)
    # off_plane_vecs = pts1 - pts4
    # return vec_angles(normals, off_plane_vecs)[0]

    # compute signed angle between the normals to the b1xb2 plane and b2xb3 plane
    b1 = pts2-pts1 # 4->1
    b2 = pts3-pts2 # 1->2
    b3 = pts4-pts3 # 2->3

    return vec_dihedrals(
        b1, b2, b3,
        crosses=crosses,
        norms=norms,
        return_crosses=return_crosses,
        **opts
    )

def pts_book_angles(pts1, pts2, pts3, pts4,
        crosses=None,
        norms=None,
        return_crosses=False,
        **opts
):
    # compute signed angle between the normals to the b1xb2 plane and b2xb3 plane
    b2 = pts1 - pts2  # 1->2
    b1 = pts2 - pts3  # 1->2
    b3 = pts4 - pts2  # 2->3

    return vec_dihedrals(
        b1, b2, b3,
        crosses=crosses,
        norms=norms,
        return_crosses=return_crosses,
        **opts
    )

################################################
#
#       mat_vec_muls
def mat_vec_muls(mats, vecs):
    """Pairwise multiplies mats and vecs

    :param mats:
    :type mats:
    :param vecs:
    :type vecs:
    :return:
    :rtype:
    """

    vecs_2 = np.matmul(mats, vecs[..., np.newaxis])
    return np.reshape(vecs_2, vecs.shape)

################################################
#
#       one_pad_vecs
def one_pad_vecs(vecs):
    ones = np.ones(vecs.shape[:-1] + (1,))
    vecs = np.concatenate([vecs, ones], axis=-1)
    return vecs

################################################
#
#       affine_multiply
def affine_multiply(mats, vecs):
    """
    Multiplies affine mats and vecs

    :param mats:
    :type mats:
    :param vecs:
    :type vecs:
    :return:
    :rtype:
    """

    vec_shape = vecs.shape
    if vec_shape[-1] != 4:
        vecs = one_pad_vecs(vecs)
    res = mat_vec_muls(mats, vecs)
    if vec_shape[-1] != 4:
        res = res[..., :3]
    return res

###
#
#       cartesian_from_rad_transforms
default_angle_sign = 1
def cartesian_from_rad_transforms(centers, vecs1, vecs2, angles, dihedrals, return_comps=False, angle_sign=None):
    """Builds a single set of affine transformation matrices to apply to the vecs1 to get a set of points

    :param centers: central coordinates
    :type centers: np.ndarray
    :param vecs1: vectors coming off of the centers
    :type vecs1: np.ndarray
    :param vecs2: vectors coming off of the centers
    :type vecs2: np.ndarray
    :param angles: angle values
    :type angles: np.ndarray
    :param dihedrals: dihedral values
    :type dihedrals: np.ndarray | None
    :return:
    :rtype:
    """
    from .TransformationMatrices import rotation_matrix, affine_matrix
    if angle_sign is None:
        angle_sign = default_angle_sign

    crosses = vec_crosses(vecs1, vecs2)
    rot_mats_1 = rotation_matrix(crosses, angle_sign*angles)
    if dihedrals is not None:
        rot_mats_2 = rotation_matrix(vecs1, -dihedrals) # add negative sign to match Gaussian sign convention
        rot_mat = np.matmul(rot_mats_2, rot_mats_1)
    else:
        rot_mat = rot_mats_1
        rot_mats_2 = None
    transfs = affine_matrix(rot_mat, centers)

    if return_comps:
        comps = (crosses, rot_mats_2, rot_mats_1)
    else:
        comps = None
    return transfs, comps

##############################################################################
#
#       cartesian_from_rad
#
def cartesian_from_rad(xa, xb, xc, r, a, d, psi=False, return_comps=False):
    """
    Constructs a Cartesian coordinate from a bond length, angle, and dihedral
    and three points defining an embedding
    :param xa: first coordinate defining the embedding
    :type xa: np.ndarray
    :param xb: third coordinate defining the embedding
    :type xb: np.ndarray
    :param xc: third coordinate defining the embedding
    :type xc: np.ndarray
    :param r:
    :type r:
    :param a:
    :type a:
    :param d:
    :type d:
    :param ref_axis:
    :type ref_axis:
    :param return_comps:
    :type return_comps:
    :return:
    :rtype:
    """

    v = xb - xa
    center = xa
    if a is None:
        vecs1 = vec_normalize(v)
        # no angle so all we have is a bond length to work with
        # means we don't even really want to build an affine transformation
        newstuff = xa + r[..., np.newaxis] * vecs1
        comps = (v, None, None, None, None)
    else:
        # print(">>>", psi)
        u = xc - xb
        if isinstance(psi, np.ndarray):
            # a = -a
            # vecs1 = vec_normalize(v)
            # v[psi] = -v[psi]
            # center = center.copy()
            # center[psi] = xb[psi]
            d = np.pi - d
            # d[psi] = np.pi-d
        # elif psi:
        #     center = xb
        #     v = xa - xb
        #     u = xa - xc
        vecs1 = vec_normalize(v)
        vecs2 = vec_normalize(u)
        transfs, comps = cartesian_from_rad_transforms(center, vecs1, vecs2, a, d,
                                                       return_comps=return_comps)
        newstuff = affine_multiply(transfs, r * vecs1)
        if return_comps:
            comps = (v, u) + comps
        else:
            comps = None
    return newstuff, comps

##############################################################################
#
#       polar_to_cartesian
#
def polar_to_cartesian_transforms(centers, vecs1, vecs2, azimuths, polars):
    """Builds a single set of affine transformation matrices to apply to the vecs1 to get a set of points

    :param centers: central coordinates
    :type centers: np.ndarray
    :param vecs1: vectors coming off of the centers
    :type vecs1: np.ndarray
    :param vecs2: vectors coming off of the centers
    :type vecs2: np.ndarray
    :param angles: angle values
    :type angles: np.ndarray
    :param dihedrals: dihedral values
    :type dihedrals: np.ndarray | None
    :return:
    :rtype:
    """
    from .TransformationMatrices import rotation_matrix, affine_matrix

    rot_mats_1 = rotation_matrix(vecs2, -azimuths)
    if polars is not None:
        vecs1 = np.broadcast_to(vecs1, rot_mats_1.shape[:-1])
        vecs2 = np.broadcast_to(vecs2, rot_mats_1.shape[:-1])
        new_ax = mat_vec_muls(rot_mats_1, vecs1)
        rot_mats_2 = rotation_matrix(vec_crosses(vecs2, new_ax), np.pi/2-polars)
        rot_mat = np.matmul(rot_mats_2, rot_mats_1)
    else:
        rot_mat = rot_mats_1
    transfs = affine_matrix(rot_mat, centers)
    return transfs

def polar_to_cartesian(center, v, u, r, a, d):
    """
    Constructs a Cartesian coordinate from a bond length, angle, and dihedral
    and three points defining an embedding
    :param xa: first coordinate defining the embedding
    :type xa: np.ndarray
    :param xb: third coordinate defining the embedding
    :type xb: np.ndarray
    :param xc: third coordinate defining the embedding
    :type xc: np.ndarray
    :param r:
    :type r:
    :param a:
    :type a:
    :param d:
    :type d:
    :param ref_axis:
    :type ref_axis:
    :param return_comps:
    :type return_comps:
    :return:
    :rtype:
    """

    if a is None:
        vecs1 = vec_normalize(v)
        # no angle so all we have is a bond length to work with
        # means we don't even really want to build an affine transformation
        newstuff = center + r[..., np.newaxis] * vecs1
    else:
        vecs1 = vec_normalize(v)
        vecs2 = vec_normalize(u)
        transfs = polar_to_cartesian_transforms(center, vecs1, vecs2, a, d)
        newstuff = affine_multiply(transfs, r[..., np.newaxis] * vecs1)
    return newstuff

##############################################################################
#
#       apply_pointwise
#
def apply_by_coordinates(tf, points, reroll=None, ndim=1, **kwargs):
    points = np.asanyarray(points)
    for i in range(ndim):
        points = np.moveaxis(points, -1, 0)

    vals = tf(*points, **kwargs)
    #TODO: this is a bit of a hack, let's clean it up
    if not isinstance(vals, np.ndarray) and isinstance(vals[0], np.ndarray):
        vals, rest = vals[0], vals[1:]
        if len(rest) == 1:
            rest = rest[0]
    else:
        rest = None

    vals = np.asanyarray(vals)
    if reroll or (reroll is None and vals.shape == points.shape):
        for i in range(ndim):
            vals = np.moveaxis(vals, 0, -1)
    if rest is not None:
        return vals, rest
    else:
        return vals

def apply_by_structures(tf, points, ndim=1, **kwargs):
    points = np.asanyarray(points)
    base_shape = points.shape[:-ndim]
    points = points.reshape((-1,) + points.shape[-ndim:])

    vals = [
        tf(pt, **kwargs)
        for pt in points
    ]


    if not (util.is_numeric(vals[0]) or util.is_numeric_array_like(vals[0])):
        vals, rest = np.asanyarray([v[0] for v in vals]), [v[1:] for v in vals]
    else:
        vals = np.asanyarray(vals)
        rest = None

    if rest is not None:
        return vals, rest
    else:
        return vals


# ##############################################################################
# #
# #       kron_sum
# #
# def kron_sum(A, B, shared=None)
#     if shared is None:
#         n = A.shape[0]
#         m = B.shape[0]
#     else:
#         raise NotImplementedError("vectorized kron sum not implemented yet...just use vec_outer tricks")
#         n = A.shape[shared]
#         m = B.shape[shared]:
#     A_ = vec_outer(A, np.eye())
#     A_ = np.kron(A, np.eye(m))
#     B_ = np.kron(np.eye(n), B)
#     if shared is not None:
#         B_ = np.moveaxis
#
#     np.moveaxis(np.kron(np.eye(), A), ) +


##############################################################################
#
#       project_out
#

def find_basis(mat, nonzero_cutoff=1e-8, method='svd'):
    if method == 'qr':
        basis, _ = np.linalg.qr(mat)
        return basis
    elif method == 'svd':
        u, s, v = np.linalg.svd(mat)
        mask = s > nonzero_cutoff
        if u.ndim == 2:
            good_s = np.where(mask)
            return u[:, good_s[0]]
        else:
            base_shape = u.shape[:-2]
            u = u.reshape((-1,) + u.shape[-2:])
            mask = mask.reshape((-1, mask.shape[-1]))
            good_sums = np.sum(mask, axis=1)
            num_good = np.unique(good_sums)
            if len(num_good) == 1:
                # num_good = num_good[0]
                # mask_inds = np.where(mask)
                # take_spec = mask_inds[:-1] + (slice(None),) + mask_inds[-1:]
                # print(take_spec)
                # u = u[take_spec].reshape(u.shape[:-1] + (num_good,))
                # return u.reshape(base_shape + u.shape[-2:])
                blocks = []
                for uu, m in zip(u, mask):
                    w = np.where(m)
                    blocks.append(uu[:, w[0]])
                blocks = np.array(blocks)
                return blocks.reshape(base_shape + blocks.shape[1:])
            else:
                blocks = []
                for uu,m in zip(u, mask):
                    w = np.where(m)
                    blocks.append(uu[:, w[0]])
                #TODO: handle base shape
                return blocks
    elif method == 'right-svd':
        u, s, v = np.linalg.svd(mat)
        mask = s > nonzero_cutoff
        if u.ndim == 2:
            good_s = np.where(mask)
            return mat @ v.T[:, good_s[0]]
        else:
            base_shape = u.shape[:-2]
            v = v.reshape((-1,) + v.shape[-2:])
            mask = mask.reshape((-1, mask.shape[-1]))
            mat = mat.reshape((-1,) + mat.shape[-2:])
            good_sums = np.sum(mask, axis=1)
            num_good = np.unique(good_sums)
            if len(num_good) == 1:
                blocks = []
                for vv, m, a in zip(v, mask, mat):
                    w = np.where(m)
                    blocks.append(a @ vv.T[:, w[0]])
                blocks = np.array(blocks)
                return blocks.reshape(base_shape + blocks.shape[1:])
            else:
                blocks = []
                for vv, m, a in zip(v, mask, mat):
                    w = np.where(m)
                    blocks.append(a @ vv.T[:, w[0]])
                # TODO: handle base shape
                return blocks
    elif method == 'right-unitary':
        u, s, v = np.linalg.svd(mat)
        mask = s > nonzero_cutoff
        if u.ndim == 2:
            good_s = np.where(mask)
            return mat @ (np.diag(1/s) @ v).T[:, good_s[0]]
        else:
            base_shape = u.shape[:-2]
            v = v.reshape((-1,) + v.shape[-2:])
            mask = mask.reshape((-1, mask.shape[-1]))
            s = s.reshape((-1, s.shape[-1]))
            mat = mat.reshape((-1,) + mat.shape[-2:])
            good_sums = np.sum(mask, axis=1)
            num_good = np.unique(good_sums)
            if len(num_good) == 1:
                blocks = []
                for vv, m, a, ss in zip(v, mask, mat, s):
                    w = np.where(m)
                    blocks.append(a @ (np.diag(1/ss) @ vv).T[:, w[0]])
                blocks = np.array(blocks)
                return blocks.reshape(base_shape + blocks.shape[1:])
            else:
                blocks = []
                for vv, m, a, ss in zip(v, mask, mat, s):
                    w = np.where(m)
                    blocks.append(a @ (np.diag(1/ss) @ vv).T[:, w[0]])
                # TODO: handle base shape
                return blocks
    else:
        raise ValueError(f"method `{method}`")

def projection_matrix(basis, orthonormal=False):
    basis = np.asanyarray(basis)
    if basis.ndim == 1:
        basis = basis[np.newaxis]
    if not orthonormal:
        basis, _ = np.linalg.qr(basis)

    return basis @ np.moveaxis(basis, -2, -1)

def orthogonal_projection_matrix(basis, orthonormal=False):
    proj = projection_matrix(basis, orthonormal=orthonormal)
    identities = identity_tensors(proj.shape[:-2], proj.shape[-1])
    return identities - proj

def _proj(projection_type, vecs, basis, ndim=None, orthonormal=False):
    vecs = np.asanyarray(vecs)
    basis = np.asanyarray(basis)
    if vecs.shape[-1] != basis.shape[-2]:
        raise ValueError(f"mismatch between vector dim {vecs.shape[-1]} and basis dim {basis.shape[-2]} ({basis.shape[-1]} basis vectors)")
    if ndim is None:
        base_shape = basis.shape[:-2]
        if len(base_shape) == 0:
            ndim = 1
        else:
            ndim = vecs.ndim - len(base_shape)
    base_shape = vecs.shape[:-ndim]

    proj = projection_type(basis, orthonormal=orthonormal)
    if ndim == 1:
        vecs = (vecs[..., np.newaxis, :] @ proj).reshape(vecs.shape)
    else:
        for _ in range(ndim):
            vecs = np.moveaxis(vecs @ proj, -1, len(base_shape))

    return vecs

def project_onto(vecs, basis, ndim=None, orthonormal=False):
    return _proj(projection_matrix, vecs, basis, ndim=ndim, orthonormal=orthonormal)

def project_out(vecs, basis, ndim=None, orthonormal=False):
    return _proj(orthogonal_projection_matrix, vecs, basis, ndim=ndim, orthonormal=orthonormal)

def fractional_power(A, pow, zero_cutoff=1e-8):
    # only applies to symmetric A
    # if symmetric:
    vals, vecs = np.linalg.eigh(A)
    if vals.ndim == 1:
        cutoffs = np.abs(vals) > zero_cutoff
        take_pos = np.where(cutoffs)[0]
        s = vals[..., take_pos]
        u = vecs[..., take_pos]
        v = np.moveaxis(u, -1, -2)
        return u @ vec_tensordiag(np.power(s, pow)) @ v
    else:
        base_shape = vals.shape[:-1]
        vals = vals.reshape((-1,) + vals.shape[-1:])
        vecs = vecs.reshape((-1,) + vecs.shape[-2:])
        cutoffs = np.abs(vals) > zero_cutoff
        cutoff_tests = np.sum(cutoffs, axis=-1)
        num_block_types = np.unique(cutoff_tests)
        blocks = len(num_block_types) > 1
        if blocks:
            take_pos = [np.where(c)[0] for c in cutoffs]
            return [
                vv[:, tp].T @ np.diag(np.power(va[tp], pow)) @ vv[:, tp]
                for vv,va,tp in zip(vecs, vals, take_pos)
            ]
        else:
            bt = num_block_types[0]
            if bt == A.shape[-1]:
                u = vecs
                v = np.moveaxis(vecs, -2, -1)
                s = vals
            else:
                take_idx = np.ravel_multi_index(np.where(cutoffs), vals.shape)
                s = vals.reshape(-1)[take_idx,].reshape((-1, bt))
                v = np.moveaxis(vecs, -2, -1).reshape((-1, vecs.shape[-1]))[
                    take_idx
                ].reshape((vecs.shape[0], bt, vecs.shape[-1]))
                u = np.moveaxis(v, -2, -1)

            flat_pow = u @ vec_tensordiag(np.power(s, pow)) @ v
            return flat_pow.reshape(base_shape + flat_pow.shape[-2:])

def unitarize_transformation(tf):
    u, s, v = np.linalg.svd(tf)
    shared_dim = min((u.shape[-1], v.shape[-2]))
    return u[..., :, :shared_dim] @ v[..., :shared_dim, :]

def polar_decomposition(tf):
    u, s, v = np.linalg.svd(tf)
    shared_dim = min((u.shape[-1], v.shape[-2]))
    Q = u[..., :, :shared_dim] @ v[..., :shared_dim, :]
    P = u @ vec_tensordiag(s) @ np.moveaxis(u, -1, -2)
    return P, Q

def maximum_similarity_transformation(basis, target, apply_transformation=True):
    lstsq_tf = np.linalg.lstsq(basis, target, rcond=None)[0]
    tf = unitarize_transformation(lstsq_tf)
    if apply_transformation:
        return basis @ tf
    else:
        return tf

def matrix_transform_from_eigs(evals, evecs, tf):
    return np.moveaxis(evecs, -1, -2) @ vec_tensordiag(tf(evals), extra_dims=evals.ndim-1) @ evecs

def symmetric_matrix_exp(mats):
    evals, evecs = np.linalg.eigh(mats)
    return matrix_transform_from_eigs(evals, evecs, np.exp)

def imaginary_symmetric_matrix_exp(mats):
    evals, evecs = np.linalg.eigh(mats)
    return [
        matrix_transform_from_eigs(evals, evecs, np.cos),
        matrix_transform_from_eigs(evals, evecs, np.sin)
    ]

def symmetric_matrix_log(mats):
    evals, evecs = np.linalg.eigh(mats)
    return matrix_transform_from_eigs(evals, evecs, np.log)

def imaginary_symmetric_matrix_log(mats_real, mats_imag):
    evals_real, evecs = np.linalg.eigh(mats_real)
    evals_imag = np.diagonal(
        evecs @ mats_imag @ np.moveaxis(evecs, -1, -2),
        axis1=-1,
        axis2=-2
    )
    return matrix_transform_from_eigs(
        np.arccos(evals_real) + np.arcsin(evals_imag), evecs, lambda x: x
    )

def sylvester_solve(A, B, C):
    A = np.asanyarray(A)
    B = np.asanyarray(B)
    C = np.asanyarray(C)
    n = A.shape[1]
    m = B.shape[0]
    S = np.kron(B.T, np.eye(n)) + np.kron(np.eye(m), A)
    val = np.linalg.solve(S, C.flatten()) # in case it can be cleverer than I can
    return val.reshape(m,n)

def symmetrize_array(a,
                     axes=None,
                     symmetrization_mode='total',
                     axes_block_ordering=None,
                     mixed_block_symmetrize=False,
                     restricted_diagonal=False,
                     out=None
                     ):
    from . import Misc as misc

    if axes is None:
        axes = np.arange(a.ndim)
    if misc.is_numeric(axes[0]):
        axes = [axes]

    axes_groups = axes
    del axes # easier debugging

    symmetrization_mode = symmetrization_mode.lower()

    if (
            not restricted_diagonal
            and axes_block_ordering is None
            and symmetrization_mode == 'total'
    ):
        if mixed_block_symmetrize:
            axes_groups = np.concatenate(axes_groups)
        b = a
        for i,axes in enumerate(axes_groups):
            a = b / math.factorial(len(axes))
            missing = np.setdiff1d(np.arange(a.ndim), axes)
            inv = np.argsort(np.concatenate([axes, missing]), axis=0)
            perm_iter = itertools.permutations(axes)
            b = a
            next(perm_iter) # first step
            if len(missing) == 0:
                for p in itertools.permutations(axes):
                    b += a.transpose(p)
            else:
                for p in itertools.permutations(axes):
                    p = np.concatenate([p, missing])[inv]
                    b += a.transpose(p)
        if out is not None:
            out[:] = b
            b = out
    else:
        flat_axes = np.concatenate(axes_groups)
        rem_axes = np.setdiff1d(np.arange(a.ndim), flat_axes)
        ordering = np.argsort(np.concatenate([flat_axes, rem_axes]))
        nmodes = a.shape[0]

        if restricted_diagonal:
            if mixed_block_symmetrize:
                pos_spec = [
                    sum(
                        [(i,)*len(axes) for i,axes in zip(term, axes_groups)],
                        ()
                    )
                    for term in itertools.combinations_with_replacement(
                        range(nmodes),
                        len(axes_groups)
                    )
                ]
            else:
                pos_spec = [
                    sum(
                        [(i,) * len(axes) for i, axes in zip(p, axes_groups)],
                        ()
                    )
                    for term in itertools.combinations_with_replacement(
                        range(nmodes),
                        len(axes_groups)
                    )
                    for p in np.unique(list(itertools.permutations(term)), axis=0)
                ]
        else:
            pos_spec = list(
                itertools.combinations_with_replacement(
                    range(nmodes),
                    len(flat_axes)
                )
            )

        if mixed_block_symmetrize:
            perms = itertools.permutations(np.arange(len(flat_axes)), len(flat_axes))
        else:
            cumlens = np.cumsum([0] + [len(a) for a in axes_groups])
            perms = (
                sum(t, ())
                for t in itertools.product(*[
                    itertools.permutations(range(l, l + len(x)), len(x))
                    for x, l in zip(axes_groups, cumlens)
                ])
            )

        if out is None:
            b = a.copy()
        else:
            b = out

        pos_arr = tuple(np.array(list(pos_spec)).T)
        padding = (slice(None),) * len(rem_axes)
        if symmetrization_mode == 'unhandled':
            for perm in perms:
                new_pos = tuple(pos_arr[flat_axes[p]] for p in perm) + padding
                new_pos = tuple(new_pos[i] for i in ordering)
                b[new_pos] = a[new_pos]
        else:
            if axes_block_ordering is not None:
                new_pos = tuple(
                    pos_arr[p] for p in
                    np.concatenate([axes_groups[f] for f in axes_block_ordering])
                ) + padding
                new_pos = tuple(new_pos[i] for i in ordering)
                val = a[new_pos]
            else:
                if symmetrization_mode == 'low':
                    new_pos = tuple(pos_arr[f] for f in reversed(flat_axes)) + padding
                    new_pos = tuple(new_pos[i] for i in ordering)
                    val = a[new_pos]
                elif symmetrization_mode == 'high':
                    new_pos = tuple(pos_arr[f] for f in flat_axes) + padding
                    new_pos = tuple(new_pos[i] for i in ordering)
                    val = a[new_pos]
                elif symmetrization_mode == 'average':
                    new_pos_1 = tuple(pos_arr[f] for f in reversed(flat_axes)) + padding
                    new_pos_1 = tuple(new_pos_1[i] for i in ordering)
                    new_pos_2 = tuple(pos_arr[f] for f in flat_axes) + padding
                    new_pos_2 = tuple(new_pos_2[i] for i in ordering)
                    val = (a[new_pos_1] + a[new_pos_2]) / 2
                else:
                    raise ValueError(
                        f"don't know what to do with `symmetrization_mode` '{symmetrization_mode}' "
                    )

            for perm in perms:
                new_pos = tuple(pos_arr[flat_axes[p]] for p in perm) + padding
                new_pos = tuple(new_pos[i] for i in ordering)
                b[new_pos] = val

    return b


import json, numbers, os

from .Jobs import ExternalProgramJob, OptionsBlock, SystemBlock

__all__ = [
    "CRESTJob"
]

class CRESTOptionsBlock(OptionsBlock):
    opts_key = None

    job_params_json = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Templates', 'crest_opts.json')
    _json = None
    @classmethod
    def load_json(cls):
        if cls._json is None:
            with open(cls.job_params_json) as opts_json:
                cls._json = json.load(opts_json)
        return cls._json

    @classmethod
    def get_props(cls):
        return list(cls.load_json()[cls.opts_key])

class CRESTProcessBlock(CRESTOptionsBlock):
    opts_key = "Process"

    def format_export(self):
        if len(self.opts) == 0:
            return None

        return "\n".join(
            f'export {k} {v}'
            for k, v in self.opts.items()
            if v is not None
        )

    def get_params(self):
        cmd = self.format_export()
        if cmd is None:
            return {}
        else:
            return {
                'command_line': cmd
            }

class CRESTCommandLineBlock(CRESTOptionsBlock):
    opts_key = "CommandLine"

    __aliases__ = {'chrg':'charge'}
    @classmethod
    def get_props(cls):
        return ["chrg"] + super().get_props()

    argument_spacer = "\\ \n" + " " * 14
    def format_command_line(self):
        if len(self.opts) == 0:
            return None
        elif len(self.opts) < 3:
            return "  ".join(
                f"--{k}" if v is True else f'--{k} {v}'
                for k, v in self.opts.items()
                if v is not None
            )
        else:
            return self.argument_spacer + self.argument_spacer.join(
                f"--{k}  " if v is True else f'--{k} {v}  '
                for k,v in self.opts.items()
                if v is not None
            )

    def get_params(self):
        cmd = self.format_command_line()
        if cmd is None:
            return {}
        else:
            return {
                'command_line': cmd
            }

class CRESTPathsBlock(OptionsBlock):
    __props__ = ('geom_file', 'crest_path', "log_file")
    __aliases__ = {'geom_file':['input_file'], 'crest_path':['path']}


class CRESTSystemBlock(SystemBlock):
    __props__ = ("atoms", "cartesians")

    fmt_key = ""

    def format_coordinate_block(self):
        atoms = self.opts.get("atoms")
        carts = self.opts.get("cartesians")
        return f"{len(atoms)}\n\n" + self.fmt_carts(atoms, carts)

    def get_params(self):
        base_opts = {}
        if len(self.opts) > 0:
            base_opts[self.fmt_key + "system"] = self.format_coordinate_block()
        return base_opts

class CRESTJob(ExternalProgramJob):
    job_template = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Templates', 'crest_job.sh')
    blocks = [
        CRESTPathsBlock,
        CRESTProcessBlock,
        CRESTCommandLineBlock,
        CRESTSystemBlock
    ]

    def __init__(self, *strs, path='crest', input_file='geom.xyz', log_file='confgen.log', **opts):
        # for o in strs:
        #     rt, o = OrcaKeywordsBlock.check_canon(o)
        #     if rt:
        #         opts[o] = True
        #     else:
        #         o[]
        opts = {o.lower():v for o,v in opts.items()}
        for o in strs:
            opts[o.lower()] = True
        super().__init__(path=path, input_file=input_file, log_file=log_file, **opts)

    # @classmethod
    # def get_extra_keys(cls):
    #     with open(cls.job_template) as r:
    #         t = r.read()
    #     return {s.strip("{").strip("}") for s in t.split()}

    @classmethod
    def get_block_types(cls):
        return cls.blocks

    @classmethod
    def load_template(cls):
        return cls.job_template

    # non_blank_line_terminated = {'link0', 'level_of_theory'}
    # def get_params(self):
    #     base_params = super().get_params()
    #     for k,b in base_params.items():
    #         if k not in self.non_blank_line_terminated:
    #             base_params[k] = b + "\n"
    #
    #     return base_params


"""
Provides hooks for viewing images, mostly through the Matplotlib Image interface
"""

__all__ = [
    "Image"
]

import numpy as np
from .Plots import ArrayPlot

class Image(ArrayPlot):
    """
    Simple subclass of ArrayPlot that just turns off most of the unnecessary features
    """

    default_opts = {
        'frame' : False,
        'ticks_style' : (False, False),#, dict(left=False, right=False))
        'ticks' : ([], []),
        'padding': ([0, 0], [0, 0]),
        'aspect_ratio':'auto'
    }
    def __init__(self, data, plot_range=None, image_size=None, **kwargs):
        if image_size is None:
            image_size = (data.shape[1], data.shape[0])
        if plot_range is None:
            plot_range = ((0, data.shape[1]), (0, data.shape[0]))
        kwargs = dict(self.default_opts, **kwargs)
        super().__init__(data, plot_range=plot_range, image_size=image_size, **kwargs)

    @classmethod
    def from_file(cls, file_name, format=None, **opts):
        import matplotlib.image as mpimg

        data = mpimg.imread(file_name, format=format)
        data = np.flip(data, axis=0)
        return cls(data, **opts)


"""
Provides basic support for point group identification and symmetry handling
"""

__all__ = []
from .Characters import *; from .Characters import __all__ as exposed
__all__ += exposed
from .Elements import *; from .Elements import __all__ as exposed
__all__ += exposed
from .Rotors import *; from .Rotors import __all__ as exposed
__all__ += exposed
from .PointGroups import *; from .PointGroups import __all__ as exposed
__all__ += exposed
from .SymmetryIdentifier import *; from .SymmetryIdentifier import __all__ as exposed
__all__ += exposed
from .Symmetrizer import *; from .Symmetrizer import __all__ as exposed
__all__ += exposed


import numpy as np

__all__ = [
    "merge_transformation_matrix"
    ]

def make_affine_matrix(mat, shift):
    """

    :param mat:
    :type mat: np.ndarray
    :param shift:
    :type shift: np.ndarray
    :return:
    :rtype:
    """
    shift = np.append(shift, [1])
    return np.column_stack(
            (
                np.row_stack(
                    (mat, np.zeros((1, 3)))
                ),
                shift
            )
        )
def merge_transformation_matrix(transf, other):
    """Merges two transformation matrices

    :param transf:
    :type transf: np.ndarray
    :param other:
    :type other: np.ndarray
    :return:
    :rtype:
    """

    other_shape = other.shape
    self_shape = transf.shape

    if other_shape[-1] == 4 and self_shape[-1] == 3:
        transf = make_affine_matrix(transf, np.array([0, 0, 0], dtype=float))
    elif other_shape[-1] == 3 and self_shape[-1] == 4:
        other = make_affine_matrix(other, np.array([0, 0, 0], dtype=float))
    elif other_shape[-1] == 3 and self_shape[-1] == 3:
        transf = make_affine_matrix(transf, np.array([0, 0, 0], dtype=float))
        other = make_affine_matrix(other, np.array([0, 0, 0], dtype=float))
    elif other_shape[-1] != 4 or self_shape[-1] != 4:
        raise ValueError("can't merge affine transforms with shape {} and {}".format(
            transf.shape,
            other.shape
        ))

    return np.dot(transf, other)



import numpy as np
import re
from ...Parsers import FileLineByLineReader

__all__ = [
    "CIFParser",
    "CIFConverter"
]

class CIFSymmetriesArray:
    def __init__(self, key, symmetry_list):
        self.key = key
        self.symms = symmetry_list
        self._arr = None
    def __repr__(self):
        cls = type(self)
        return f"{cls.__name__}(<{len(self.symms)}>)"

    def _parse_symmetry_function(self, arr, row, sf):
        sums = sf.split("+")
        for sm_cmp in sums:
            if len(sm_cmp) == 0: continue
            for i,sub_cmp in enumerate(sm_cmp.split("-")):
                if len(sub_cmp) == 0: continue
                # positive term is the first, the rest are negative
                parity = (-1) if i > 0 else 1
                scaling_comps = sub_cmp.split('*')
                if len(scaling_comps) == 1:
                    scaling = 1
                else:
                    if '/' in scaling_comps[0]:
                        num, den = scaling_comps[0].split('/')
                        scaling = float(num) / float(den)
                    else:
                        scaling = float(scaling_comps[0])
                comp = scaling_comps[-1]
                idx = {'x':0,'y':1,'z':2}.get(comp)
                if idx is None:
                    if '/' in comp:
                        num, den = comp.split('/')
                        comp = float(num) / float(den)
                    else:
                        comp = float(comp)
                    arr[row, 3] = scaling * float(comp) * parity
                else:
                    arr[row, idx] = scaling * parity
    def _get_symmetries(self, key, symm_labels):
        if isinstance(symm_labels, str):
            arr = np.zeros((4, 4)) # affine transformation
            arr[-1, -1] = 1
            cols = symm_labels.split(",")
            for i,c in enumerate(cols):
                self._parse_symmetry_function(arr, i, c)
            return arr
        else:
            return np.array([self._get_symmetries(key, sf) for sf in symm_labels])
    @property
    def transformation(self):
        if self._arr is None:
            self._arr = self._get_symmetries(self.key, self.symms)
        return self._arr

class CIFParser(FileLineByLineReader):
    # tags = {'_database_code_CSD', '_cell_length_a', '_cell_length_b', '_cell_length_c',
    #         '_cell_angle_alpha', '_cell_angle_beta', '_cell_angle_gamma',
    #         '_cell_formula_units_Z', '_chemical_formula_moiety', '_chemical_formula_sum',
    #         '_atom_site_fract_z', 'atom_type_radius_bond'}
    # aliases = {
    #     '_database_code_CSD': 'name', '_cell_length_a': 'a', '_cell_length_b': 'b', '_cell_length_c': 'c',
    #     '_cell_angle_alpha': 'alpha', '_cell_angle_beta': 'beta',
    #     '_cell_angle_gamma': 'gamma', '_cell_formula_units_Z': 'Z',
    #     '_chemical_formula_sum': 'sum', '_chemical_formula_moiety': 'moiety',
    #     '_atom_site_fract_z': 'atomloop', 'atom_type_radius_bond': 'bondradii'
    # }
    # exceptions = {'_chemical_formula_moiety', '_atom_site_fract_z', 'atom_type_radius_bond'}

    def __init__(self, file, fields=None, **kw):
        super().__init__(file, max_nesting_depth=1, **kw)
        self.fields = fields
    def check_tag(self, line:str, depth:int=0, active_tag=None, label:str=None, history:list[str]=None):
        if len(line) == 0:
            return self.LineReaderTags.SKIP
        elif line.startswith('data_'):
            return self.LineReaderTags.BLOCK_START, line[5:], None
        elif line.startswith('#'):
            if line == '#END':
                return self.LineReaderTags.BLOCK_END
            else:
                return self.LineReaderTags.COMMENT, line.lstrip('#')
        elif line.startswith('loop_'):
            return self.LineReaderTags.BLOCK_START
        elif line.startswith('_'):
            in_loop = depth == 1 and active_tag is self.LineReaderTags.BLOCK_START and label is None
            if in_loop:
                if len(history) > 0 and isinstance(history[-1], str) and not history[-1].startswith('_'):
                    return self.LineReaderTags.RESETTING_BLOCK_END
                else:
                    return None
            else:
                bits = line[1:].split(maxsplit=1)
                if len(bits) == 1: bits = bits + [""]
                return [self.LineReaderTags.BLOCK_START] + bits
        else:
            return None

    custom_handlers = {}
    def get_block_handlers(self):
        return {
            'cell_length_a':self._get_float,
            'cell_length_b':self._get_float,
            'cell_length_c':self._get_float,
            'cell_angle_alpha':self._get_float,
            'cell_angle_beta':self._get_float,
            'cell_angle_gamma':self._get_float,
            'cell_volume':self._get_float,
            'journal_page_first':self._get_int,
            'journal_page_last':self._get_int,
            'journal_volume':self._get_int,
            'journal_year':self._get_int,
            'atom_site_symmetry_multiplicity':self._get_int,
            'atom_site_fract_x':self._get_float,
            'atom_site_fract_y':self._get_float,
            'atom_site_fract_z':self._get_float,
            'atom_site_occupancy':self._get_float,
            'atom_type_oxidation_number':self._get_float,
            'symmetry_equiv_pos_as_xyz':CIFSymmetriesArray
        }
    def _get_float(self, key, val):
        if isinstance(val, str):
            return float(val)
        else:
            return np.array(val).astype(float)
    def _get_int(self, key, val):
        if isinstance(val, str):
            return float(val)
        else:
            return np.array(val).astype(int)
    def resolve_handler(self, label:'str|None'):
        handler = self.get_block_handlers().get(label)
        if label is not None:
            if label.endswith('_num') or label.endswith('_number'):
                handler = self._get_int
        return handler

    def handle_block(self, label:'str|None', block_data, join=True, depth=0):
        handler = self.get_block_handlers().get(label)
        if handler is not None:
            if len(block_data) == 1:
                return handler(label, block_data[0])
            else:
                return handler(label, block_data)
        else:
            if label is not None:
                if not join or isinstance(block_data, str) or not all(isinstance(b, str) for b in block_data):
                    return block_data
                else:
                    return "".join(block_data).strip()
            else:
                key_list = []
                datasets = {}
                key_parse = True
                for line in block_data:
                    if key_parse and line.startswith("_"):
                        key_list.append(line[1:].strip())
                    else:
                        if key_parse:
                            key_parse = False # keys come first done
                            for k in key_list:
                                datasets[k] = []
                        for k,v in zip(key_list, line.split()):
                            datasets[k].append(v)
                return {k:self.handle_block(k, v, join=False) for k,v in datasets.items()}

    # MAX_BLOCKS = 10
    def parse(self, target_fields=None):
        tmp_fields = self.fields
        try:
            self.fields = target_fields
            return list(iter(self))
        finally:
            self.fields = tmp_fields

class CIFConverter:
    def __init__(self, parsed_cif):
        self.data = parsed_cif
        self._cached_queries = {}
    @property
    def cell_properties(self):
        return self.prep_property_dict(
            self.find_all('cell_*', strict=False, cache=True)
        )
    @property
    def atom_properties(self):
        return self.prep_property_dict(
            self.find_all('atom_*', strict=False, cache=True)
        )
    @property
    def symmetry_properties(self):
        return self.prep_property_dict(
            self.find_all('symmetry_*', strict=False, cache=True)
        )
    def prep_property_dict(self, res):
        return {
            k:v
            for d in res
            for k,v in d.items()
        }
    def find(self, item, strict=True, cache=False):
        key = (item, strict)
        if key in self._cached_queries:
            return self._cached_queries[key]
        res = None
        if strict:
            for a in self.data:
                if item in a:
                    res = a[item]
                    break
        else:
            pattern = re.compile(item)
            for a in self.data:
                for k,v in a.items():
                    if re.match(pattern, k):
                        res = {k:v}
                        break
                if res is not None:
                    break
        if cache:
            self._cached_queries[key] = res
        return res
    def find_all(self, item, strict=True, cache=False):
        key = (item, strict)
        if key in self._cached_queries:
            return self._cached_queries[key]
        if strict:
            res = [
                a[item] for a in self.data if item in a
            ]
        else:
            pattern = re.compile(item)
            filter = [
                {k:v for k,v in a.items() if re.match(pattern, k)}
                for a in self.data
            ]
            res = [f for f in filter if len(f) > 0]
        if cache:
            self._cached_queries[key] = res
        return res

    @property
    def atoms(self):
        return self.construct_atom_coords(
            self.atom_properties,
            self.symmetry_properties
        )
    def construct_base_atom_coords(self, property_dict):
        elems_charges = [
            (a[:2], int(a[2:]))
                if not a[1].isdigit() else
            (a[:1], int(a[1:]))
            for a in property_dict['atom_site_type_symbol']
        ]
        coords = np.array([
            property_dict.get(f'atom_site_fract_{x}')
            for x in ['x', 'y', 'z']
        ]).T
        return {
            'atoms':[e for e,c in elems_charges],
            'coords':coords,
            'charges':[c for e,c in elems_charges]
        }
    def construct_atom_coords(self, atom_properties, symmetry_properties):
        structs = self.construct_base_atom_coords(atom_properties)
        symm_ops = symmetry_properties.get('symmetry_equiv_pos_as_xyz')
        if symm_ops is not None:
            ats = structs['atoms']
            crd = structs['coords']
            chg = structs['charges']
            crd = np.expand_dims(
                np.concatenate([crd, np.ones((crd.shape[0], 1))], axis=1),
                [0, 3]
            )
            all_symm = np.expand_dims(symm_ops.transformation, 1)
            full_crd = all_symm @ crd
            crd = full_crd[:, :, :3, 0].reshape(-1, 3)
            ats = list(ats) * full_crd.shape[0]
            chg = list(chg) * full_crd.shape[0]
            structs = {
                'atoms':ats,
                'coords':crd,
                'charges':np.array(chg)
            }
        return structs




from .CommonData import DataHandler, DataError

__all__ = [ "BondData", "BondDataHandler" ]
__reload_hook__ = [".CommonData"]

class BondDataHandler(DataHandler):
    """
    A DataHandler that's built for use with the bond data we've collected.
    Usually used through the `BondData` object.
    """
    def __init__(self):
        super().__init__("BondData", getter=lambda d,k:self.get_distance(k))
    def load(self):
        super().load()
        # now we create a back mapping
        for k in tuple(self._data.keys()):
            sub = self._data[k]
            for a in sub.keys():
                if a not in self._data:
                    self._data[a] = {}
                da = self._data[a]
                for n,v in sub[a].items():
                    if k not in da:
                        da[k] = {n:v}
                    else:
                        da[k][n] = v
    def get_distance(self, key, default=None):
        if isinstance(key, str):
            a1 = key
            a2 = None
            w = None
        elif len(key) == 1:
            a1 = key[0]
            a2 = None
            w = None
        else:
            try:
                a1, a2, w = key
            except (TypeError, ValueError):
                a1, a2 = key
                w = "first"

        d = self.data
        if a1 not in d:
            if default is None:
                raise DataError("Bond data for {} is unknown".format(a1))
            else:
                return default
        d = d[a1]
        if a2 is None:
            return d
        if a2 not in d:
            if default is None:
                raise DataError("Bond data for a {}-{} bond is unknown".format(a1, a2))
            else:
                return default
        d = d[a2]

        if w is None:
            return d
        if isinstance(w, int):
            if w == 1:
                w = "Single"
            elif w == 2:
                w = "Double"
            elif w == 3:
                w = "Triple"
        if isinstance(w, str):
            if w == "first":
                ww = None
                dd = d.get("Single")
                if dd is None:
                    dd = d.get("Double")
                else:
                    ww = "Single"

                if ww is None:
                    if dd is None:
                        dd = d.get("Triple")
                    else:
                        ww = "Double"

                if ww is None:
                    if dd is None:
                        ww = next(iter(d.keys()))
                    else:
                        ww = "Triple"
                w = ww
            else:
                w = w.capitalize()

        if w not in d:
            if default is None:
                raise DataError("Bond data for a {}-{} {} bond is unknown".format(a1, a2, w))
            else:
                return default

        return d[w]

    def __getitem__(self, item):
        """
        Special cases the default getitem so tuples are mapped
        :param item:
        :type item:
        :return:
        :rtype:
        """
        if isinstance(item, tuple):
            return self.get_distance(item)
        else:
            return super().__getitem__(item)

BondData=BondDataHandler()
BondData.__doc__ = """An instance of BondDataHandler that can be used for looking up bond distances"""
BondData.__name__ = "BondData"


import abc
import uuid

import numpy as np

from ..Jupyter import JHTML, X3DHTML
from .. import Numputils as nput

__all__ = [
    "X3D",
    "X3DPrimitive",
    "X3DGeometryObject",
    "X3DGeometryGroup",
    "X3DGroup",
    "X3DScene",
    "X3DBackground",
    "X3DMaterial",
    "X3DLine",
    "X3DSphere",
    "X3DCone",
    "X3DCylinder",
    "X3DCappedCylinder",
    "X3DArrow",
    "X3DTorus",
    "X3DRectangle2D",
    "X3DDisk2D",
    "X3DCircle2D",
    "X3DPolyline2D",
    "X3DTriangleSet",
    "X3DIndexedTriangleSet",
    "X3DIndexedLineSet",
    "X3DSwitch",
    "X3DListAnimator"
]

class X3DObject(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def to_x3d(self):
        ...

class X3D(X3DObject):
    defaults = dict(
        width=500,
        height=500
    )
    def __init__(self, *children, id=None, dynamic_loading=True, **opts):
        if len(children) == 1 and isinstance(children[0], (tuple, list)):
            children = children[0]
        self.children = children
        self.opts = opts
        if id is None:
            id = "x3d-" + str(uuid.uuid4())[:6]
        self.id = id
        self.dynamic_loading = dynamic_loading

    X3DOM_JS = 'https://www.x3dom.org/download/1.8.3/x3dom-full.js'
    X3DOM_CSS = 'https://www.x3dom.org/download/x3dom.css'

    def get_export_script(self, id):
        return f"""
(function(){{
  let link = document.createElement('a');
  let base_name = '{id}';
  link.download = base_name + '.png';
  link.href = document.getElementById('{id}').getElementsByTagName('canvas')[0].toDataURL()
  link.click();
}})()
       """

    def get_record_screen_script(self, id, polling_rate=30, recording_duration=2, video_format='video/webm'):
            return f"""
    (function(){{
        let canvas = document.getElementById('{id}').getElementsByTagName('canvas')[0];
        
        let pollingRate = (typeof canvas.pollingRate === 'undefined') ? {polling_rate} : canvas.pollingRate;
        let videoFormat = (typeof canvas.videoFormat === 'undefined') ? "{video_format}" : canvas.videoFormat;
        let videoExtension = canvas.videoExtension;
        if (typeof canvas.videoExtension === 'undefined') {{
            videoExtension = ""
        }}
        let x3DRecordingStream = canvas.captureStream(pollingRate);
        let mediaRecorder = new MediaRecorder(x3DRecordingStream, {{mimeType: videoFormat}});
        
        mediaRecorder.frames = [];
        mediaRecorder.ondataavailable = function(e) {{
          mediaRecorder.frames.push(e.data);
        }};
        
        mediaRecorder.onstop = function(e) {{
          link = document.createElement('a');
          const base_name = '{id}';
          const blob = mediaRecorder.frames[0];
          link.download = base_name + videoExtension;
          console.log(blob);
          const blobURL = window.URL.createObjectURL(blob);
          link.href = blobURL;
          console.log(blobURL);
          mediaRecorder.frames = [];
          link.click();
        }};
        
        let duration = (typeof canvas.recordingDuration === 'undefined') ? {recording_duration} : canvas.recordingDuration;
        setTimeout(() => {{mediaRecorder.stop()}}, duration * 1000);
        mediaRecorder.start()
    }})()
           """

    def set_animation_duration_script(self, id):
        return f"""
    (function(){{
        let canvas = document.getElementById('{id}').getElementsByTagName('canvas')[0];
        let input = document.getElementById('{id}-input');
        
        canvas.recordingDuration = input.value;
    }})()
           """

    include_export_button = False
    include_record_button = False
    def to_widget(self, dynamic_loading=None, include_export_button=None, include_record_button=None):
        id = self.id
        x3d_embed = self.to_x3d()#.tostring()

        if dynamic_loading is None:
            dynamic_loading = self.dynamic_loading
        if include_export_button is None:
            include_export_button = self.include_export_button
        if include_record_button is None:
            include_record_button = self.include_record_button

        if not dynamic_loading:
            base_fig = JHTML.Div(
                JHTML.Link(rel='stylesheet', href=self.X3DOM_CSS),
                JHTML.Script(src=self.X3DOM_JS),
                x3d_embed,
                id=id
            )
        else:
            JHTML.Link(rel='stylesheet', href=self.X3DOM_CSS),
            load_script = JHTML.Script(src=self.X3DOM_JS).tostring()
            kill_id = "tmp-"+str(uuid.uuid4())[:10]
            base_fig = JHTML.Figure(
                # JHTML.Link(rel='stylesheet', href=self.X3DOM_CSS),
                x3d_embed,
                JHTML.Image(
                    src='data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7',
                    id=kill_id,
                    onload=f"""
                    (function() {{
                        document.getElementById("{kill_id}").remove();
                        const frag = document.createRange().createContextualFragment(`{load_script}`);
                        document.head.appendChild(frag);
                    }})()"""
                    ),
                id=id
            )

        elems = [base_fig]
        if include_export_button:
            elems.append(JHTML.Button("Save Figure", onclick=self.get_export_script(self.id)))
        if include_record_button:
            elems.extend([
                JHTML.Button("Record Animation", onclick=self.get_record_screen_script(self.id)),
                JHTML.Input(value="2", id=self.id+'-input', oninput=self.set_animation_duration_script(self.id))
            ])

        if len(elems) > 1:
            return JHTML.Div(
                *elems
            )
        else:
            return elems[0]

    def to_html(self, *base_elems, header_elems=None,
                dynamic_loading=False,
                include_export_button=None,
                include_record_button=None,
                **header_info):
        id = self.id
        x3d_embed = self.to_widget(
            dynamic_loading=dynamic_loading,
            include_export_button=include_export_button,
            include_record_button=include_record_button
        )  # .tostring()

        return JHTML.Html(
            JHTML.Head(
                *(header_elems if header_elems is not None else []),
                JHTML.Link(rel='stylesheet', href=self.X3DOM_CSS),
                JHTML.Script(src=self.X3DOM_JS),
                **header_info
            ),
            JHTML.Body(
                *base_elems,
                x3d_embed
            ),
            id=id
        )

    def _ipython_display_(self):
        return self.to_widget()._ipython_display_()
    def get_mime_bundle(self):
        return self.to_widget().get_mime_bundle()
    def to_x3d(self):
        base_opts = dict(self.defaults, **self.opts)
        for k in ['width', 'height']:
            if k in base_opts:
                v = base_opts[k]
                if nput.is_numeric(v):
                    base_opts[k] = f'{v:.0f}px'
        return X3DHTML.X3D(
            JHTML.HTML.Head(),
            *[a.to_x3d() if hasattr(a, 'to_x3d') else a for a in self.children],
            **base_opts
        )
    def display(self):
        return self.to_widget().display()

    def show(self):
        from ..Jupyter.JHTML import JupyterAPIs
        dynamic_loading = JupyterAPIs().in_jupyter_environment()
        self.to_widget(dynamic_loading=dynamic_loading).display()

    def dump(self, file, write_html=True, **opts):
        if write_html:
            html = self.to_html()
        else:
            html = self.to_x3d()
        return html.write(file, **opts)

class X3DOptionsSet(X3DObject):
    __props__ = {}

    def __init__(self, **attrs):
        self.attrs = attrs

    conversion_map = {}
    @classmethod
    def prop_keys(cls):
        return (cls.__props__ | cls.conversion_map.keys())
    def prep_attrs(self, attrs:dict):
        attrs = {
            self.conversion_map.get(k, k):v
            for k,v in attrs.items()
        }
        excess_keys = attrs.keys() - self.__props__
        if len(excess_keys) > 0:
            cls = type(self).__name__
            raise ValueError(f"keys {excess_keys} are invalid keys for {cls}")
        return attrs

class X3DMaterial(X3DOptionsSet):
    __props__ = {
        "diffuseColor",
        "ambientIntensity",
        "emissiveColor",
        "specularColor",
        "shininess",
        "transparency",
        # "metadata"
    }
    conversion_map = {
        "brightness": "ambientIntensity",
        "glow": "emissiveColor",
        "color": "diffuseColor",
        "specularity": "specularColor"
    }
    def to_x3d(self):
        return X3DHTML.Material(**self.prep_attrs(self.attrs))

class X3DAppearance(X3DOptionsSet):
    __props__ = {
        "alphaClipThreshold",
        "blendMode",
        "colorMaskMode"
        "depthMode"
        "lineProperties"
        "material"
        "metadata"
        "pointProperties"
        "shaders",
        "sortKey",
        "sortType",
        "texture",
        "textureTransform"
    }
    def prep_attrs(self, attrs:dict):
        material_keys = attrs.keys() & X3DMaterial.prop_keys()
        line_keys = attrs.keys() & X3DLineProperties.prop_keys()
        point_keys = attrs.keys() & X3DPointProperties.prop_keys()

        rem_keys = attrs.keys() - (
            X3DMaterial.prop_keys()
            | X3DLineProperties.prop_keys()
            | X3DPointProperties.prop_keys()
        )
        base_attrs = {k:attrs[k] for k in rem_keys}

        if len(material_keys) > 0:
            base_attrs['material'] = {k:attrs[k] for k in material_keys}

        if len(line_keys) > 0:
            base_attrs['lineProperties'] = {k:attrs[k] for k in line_keys}

        if len(point_keys) > 0:
            base_attrs['pointProperties'] = {k:attrs[k] for k in point_keys}

        return base_attrs
    def to_x3d(self):
        base_attrs = self.prep_attrs(self.attrs)

        comps = []
        line_props = base_attrs.pop('lineProperties', None)
        if line_props is not None:
            if isinstance(line_props, dict):
                line_props = X3DLineProperties(**line_props)
            if isinstance(line_props, X3DLineProperties):
                line_props = line_props.to_x3d()
            comps.append(line_props)

        point_props = base_attrs.pop('pointProperties', None)
        if point_props is not None:
            if isinstance(point_props, dict):
                point_props = X3DPointProperties(**point_props)
            if isinstance(point_props, X3DPointProperties):
                point_props = point_props.to_x3d()
            comps.append(point_props)

        material_props = base_attrs.pop('material', None)
        if material_props is not None:
            if isinstance(material_props, dict):
                material_props = X3DMaterial(**material_props)
            if isinstance(material_props, X3DMaterial):
                material_props = material_props.to_x3d()
            comps.append(material_props)

        return X3DHTML.Appearance(*comps, **base_attrs)

class X3DLineProperties(X3DOptionsSet):
    __props__ = {
        "applied",
        "linetype",
        "linewidth",
        "linewidthScaleFactor"
    }
    conversion_map = {
        "line_style":"linetype",
        "line_thickness":"linewidth"
    }
    def prep_attrs(self, attrs:dict):
        attrs = super().prep_attrs(attrs)
        attrs['linewidthScaleFactor'] = attrs.get('linewidthScaleFactor', '1')
        attrs['containerField'] = attrs.get('containerField', 'lineProperties')
        return attrs
    def to_x3d(self):
        return X3DHTML.LineProperties(**self.prep_attrs(self.attrs))

class X3DPointProperties(X3DOptionsSet):
    __props__ = {
        "attenuation",
        "pointSizeMaxValue",
        "pointSizeMinValue",
        "pointSizeScaleFactor"
    }
    conversion_map = {
        "point_size":"pointSizeScaleFactor"
    }
    def prep_attrs(self, attrs:dict):
        attrs = super().prep_attrs(attrs)
        attrs['containerField'] = attrs.get('containerField', 'pointProperties')
        attrs['pointSizeMaxValue'] = str(
            max([
                float(attrs.get('pointSizeMaxValue', '0')),
                float(attrs.get('pointSizeMinValue', '0')),
                float(attrs.get('pointSizeScaleFactor', '0')),
            ])
        )
        return attrs
    def to_x3d(self):
        return X3DHTML.PointProperties(**self.prep_attrs(self.attrs))

class X3DPrimitive(X3DObject):
    wrapper_class = None
    tag_class = None
    def __init__(self, *children, **opts):
        if len(children) == 1 and isinstance(children[0], (tuple, list)):
            children = children[0]
        self.children = children
        self.opts = opts
    def split_opts(self, opts:dict):
        material_keys = opts.keys() & (
            X3DMaterial.prop_keys()
            | X3DAppearance.prop_keys()
            | X3DLineProperties.prop_keys()
            | X3DPointProperties.prop_keys()
        )
        rem_keys = opts.keys() - material_keys
        return {k:opts[k] for k in rem_keys}, {k:opts[k] for k in material_keys}
    def get_appearance(self, appearance_options):
        if len(appearance_options) > 0:
            return X3DAppearance(**appearance_options).to_x3d()
        else:
            return None
    def to_x3d(self):
        obj_opts, appearance_opts = self.split_opts(self.opts)
        kids = [k.to_x3d() if hasattr(k, 'to_x3d') else k for k in self.children]
        appearance = self.get_appearance(appearance_opts)
        if self.tag_class is None:
            if appearance is not None:
                kids = [appearance] + kids
            return self.wrapper_class(
                kids,
                **obj_opts
            )
        else:
            core = self.tag_class(kids, **obj_opts)
            appearance = self.get_appearance(appearance_opts)
            if appearance is not None:
                return self.wrapper_class(appearance, core)
            else:
                return core

class X3DScene(X3DPrimitive):
    wrapper_class = X3DHTML.Scene
    default_viewpoint = {'viewAll':True}
    children: list
    def __init__(self, *children:X3DPrimitive, background=None, viewpoint=None, **opts):
        if viewpoint is None:
            viewpoint = self.default_viewpoint
        elif viewpoint is False:
            viewpoint = {}
        super().__init__(*children, **opts)
        if background is not None:
            self.children = [X3DBackground(color=background)] + list(self.children)
        if len(viewpoint) > 0:
            viewpoint = self.get_view_settings(**viewpoint)
            self.children = [X3DHTML.Viewpoint(**viewpoint)] + list(self.children)

    default_up_vector = (0, 1, 0)
    default_right_vector = (1, 0, 0)
    default_view_vector = (0, 0, 1)
    default_view_distance = 10
    @classmethod
    def get_view_settings(cls,
                          up_vector=None, view_vector=None, right_vector=None,
                          view_distance=None, view_center=None, view_matrix=None, view_position=None, **etc):
        # CO = coords0[1] - coords0[0]
        # OH = coords0[5] - coords0[1]
        if view_matrix is None:
            if view_vector is None:
                if (
                    up_vector is not None and right_vector is not None
                ):
                    view_vector = nput.vec_crosses(up_vector, right_vector, normalize=True)
                elif right_vector is not None:
                    view_vector = nput.vec_crosses(cls.default_up_vector, right_vector, normalize=True)
                elif up_vector is not None:
                    view_vector = nput.vec_crosses(up_vector, cls.default_right_vector, normalize=True)

            if view_vector is not None:
                m = nput.rotation_matrix(
                    view_vector,
                    cls.default_view_vector
                )
            else:
                m = np.eye(3)

            if up_vector is None and right_vector is not None:
                if view_vector is None:
                    view_vector = cls.default_view_vector
                up_vector = nput.vec_normalize(
                    nput.vec_crosses(view_vector, right_vector)
                )
            if up_vector is not None:
                m = m @ nput.rotation_matrix(
                    m.T @ up_vector,
                    cls.default_up_vector
                )
            view_matrix = m

        ang, cross = nput.extract_rotation_angle_axis(view_matrix)
        if view_vector is None:
            view_vector = view_matrix[:, -1]
        if view_position is None:
            if view_distance is None:
                view_distance = cls.default_view_distance
            view_position = view_distance * nput.vec_normalize(np.asanyarray(view_vector))
            if view_center is not None:
                view_center = view_matrix @ np.asanyarray(view_center)
                view_position = view_distance * nput.vec_normalize(
                    view_position + view_center
                )
        else:
            view_position = view_matrix @ np.asanyarray(view_position)
        return dict(
            {
                'orientation': list(cross) + [ang],
                'position': view_position
            },
            **etc
        )

class X3DBackground(X3DOptionsSet):
    wrapper_class = X3DHTML.Background
    __props__ = {
        'skyColor',
        'skyAngle'
    }
    conversion_map = {
        "color": "skyColor"
    }
    def prep_attrs(self, attrs: dict):
        attrs = super().prep_attrs(attrs)
        color = attrs.get('skyColor', None)
        if color is not None:
            if isinstance(color, (list, tuple, np.ndarray)) and all(isinstance(c, str) for c in color):
                color = " ".join(color)
            if isinstance(color, str):
                try:
                    _ = [float(s) for s in color.split()]
                except ValueError:
                    from .Colors import ColorPalette

                    bits = []
                    for c in color.split():
                        if color.startswith('#'):
                            c = ColorPalette.parse_rgb_code(c)
                        else:
                            c = np.array(ColorPalette.parse_color_string(c))
                        bits.extend(np.array(c) / 255)
                    attrs['skyColor'] = bits
        return attrs

    def to_x3d(self):
        return X3DHTML.Background(**self.prep_attrs(self.attrs))

class X3DCoordinate(X3DPrimitive):
    wrapper_class = X3DHTML.Coordinate
    def __init__(self, points):
        super().__init__(point=self.prep_points(points))
    @classmethod
    def prep_points(cls, points):
        return " ".join(np.asanyarray(np.round(points, 4)).flatten().astype(str))

class X3DColor(X3DPrimitive):
    wrapper_class = X3DHTML.Color
    def __init__(self, colors):
        super().__init__(color_list=colors)
    def split_opts(self, opts:dict):
        base_opts, appearance_opts = super().split_opts(opts)
        base_opts['color'] = self.prep_color(base_opts.pop('color_list'))
        return base_opts, appearance_opts
    @classmethod
    def prep_color(cls, points):
        return " ".join(np.asanyarray(np.round(points, 4)).flatten().astype(str))

class X3DGroup(X3DPrimitive):
    wrapper_class = X3DHTML.Group

class X3DSwitch(X3DPrimitive):
    wrapper_class = X3DHTML.Switch

class X3DGeometryObject(X3DPrimitive):
    wrapper_class = X3DHTML.Shape
    def __init__(self, *args, **opts):
        geom_opts, self.material_opts = self.split_opts(opts)
        self.geometry_opts = self.prep_geometry_opts(*args, **geom_opts)
        super().__init__()
    @abc.abstractmethod
    def prep_geometry_opts(self, *args, **opts) -> dict:
        ...
    def create_tag_object(self, **core_opts):
        return self.tag_class(**core_opts)
    def create_object(self,
                      translation=None,
                      rotation=None,
                      scale=None,
                      normal=None,
                      up_vector=None,
                      **core_opts):
        base_obj = self.create_tag_object(**core_opts)
        tf = {}
        if normal is not None:
            if up_vector is None:
                up_vector = [0, 0, 1]
            angs, crosses = nput.vec_angles(up_vector, normal, return_crosses=True, return_norms=False)
            if rotation is not None:
                if isinstance(rotation, str):
                    rotation = np.array(rotation.split()).astype(float)
                angs, crosses = nput.extract_rotation_angle_axis(
                    nput.rotation_matrix(crosses, angs)
                        @ nput.rotation_matrix(rotation[:3], rotation[3])
                )
            rotation = np.concatenate([crosses, [angs]])
        for k,v in [["translation",translation], ["rotation",rotation], ["scale", scale]]:
            if v is not None:
                tf[k] = np.round(v, 4) if not isinstance(v, str) else v
        if len(tf) == 0:
            tf = None
        # base_obj = X3DHTML.Transform(base_obj, translation=translation, rotation=rotation, scale=scale)
        return base_obj, tf
    def get_rotation(self, axis, up_vector=None):
        if up_vector is None:
            up_vector = [0, 1, 0]
        angs, crosses, norms = nput.vec_angles(up_vector, axis, return_crosses=True, return_norms=True)
        if nput.is_numeric(angs):
            return np.concatenate([crosses, [angs]]), norms[1]
        else:
            return np.concatenate([crosses, angs[..., np.newaxis]], axis=-1), norms[1]
    def to_x3d(self):
        # obj_opts, material_opts = self.split_opts(self.opts)
        # kids = [k.to_x3d() for k in self.children]
        core, tf = self.create_object(**self.geometry_opts)
        appearance = self.get_appearance(self.material_opts)
        if appearance is not None:
            core = self.wrapper_class(appearance, core)
        if tf is not None:
            core = X3DHTML.Transform(core, **tf)
        return core

class X3DGeometryGroup(X3DGeometryObject):
    @abc.abstractmethod
    def prep_geometry_opts(self, *args, **opts) -> list[dict]:
        ...
    def prep_vecs(self, vecs, nstruct=None):
        if vecs is None:
            return [None] * nstruct
        else:
            vecs = np.asanyarray(vecs)
            if vecs.ndim == 1:
                vecs = vecs[np.newaxis]
            if nstruct is not None:
                vecs = np.broadcast_to(vecs, (nstruct, vecs.shape[-1]))
        return vecs
    def prep_mats(self, mats, nstruct=None):
        if mats is None:
            return [None] * nstruct
        else:
            mats = np.asanyarray(mats)
            if mats.ndim == 2:
                mats = mats[np.newaxis]
            if nstruct is not None:
                mats = np.broadcast_to(mats, (nstruct,) + mats.shape[-2:])
            return mats
    def prep_const(self, const, nstruct):
        if const is None:
            return [None] * nstruct
        else:
            const = np.asanyarray(const)
            if const.ndim == 0:
                const = const[np.newaxis]
            const = np.broadcast_to(const, (nstruct,))
            return const
    def to_x3d(self):
        kids = [self.create_object(**g) for g in self.geometry_opts]
        appearance = self.get_appearance(self.material_opts)
        objs = []
        for o,tf in kids:
            if appearance is not None:
                o = self.wrapper_class(appearance, o)
            if tf is not None:
                o = X3DHTML.Transform(o, **tf)
            objs.append(o)
        if len(objs) == 1:
            return objs[0]
        else:
            return X3DHTML.Group(objs)

class X3DSphere(X3DGeometryGroup):
    tag_class = X3DHTML.Sphere

    def prep_geometry_opts(self, centers, radius=1, **opts):
        centers = self.prep_vecs(centers)
        rads = self.prep_const(radius, centers.shape[0])
        return [{"translation":c, "radius":r} for c,r in zip(centers, rads)]

class X3DCylinder(X3DGeometryGroup):
    tag_class = X3DHTML.Cylinder

    def prep_geometry_opts(self, starts, ends, radius=1, **opts):
        starts = self.prep_vecs(starts)
        ends = self.prep_vecs(ends)
        radius = self.prep_const(radius, starts.shape[0])

        axes = ends - starts
        rots, norms = self.get_rotation(axes)

        return [
            {"translation":s, "rotation":a, "height":n, "radius":r}
            for s,a,n,r in zip((starts + ends) / 2, rots, norms, radius)
        ]

class X3DCone(X3DGeometryGroup):
    tag_class = X3DHTML.Cone

    def prep_geometry_opts(self, starts, ends, radius=1, top_radius=None, **opts):
        starts = self.prep_vecs(starts)
        ends = self.prep_vecs(ends)
        radius = self.prep_const(radius, starts.shape[0])
        top_radius = self.prep_const(top_radius, starts.shape[0])

        axes = ends - starts
        rots, norms = self.get_rotation(axes)

        return [
            {"translation":s, "rotation":a, "height":n, "bottomRadius":r, "topRadius":t}
            for s,a,n,r,t in zip((starts + ends) / 2, rots, norms, radius, top_radius)
        ]

class X3DArrow(X3DGroup):

    arrowhead_class = X3DCone
    cylinder_class = X3DCylinder

    def __init__(self,
                 starts, ends,
                 radius=1,
                 top_radius=None,
                 arrowhead_radius=2,
                 arrowhead_radius_mode='scaled',
                 arrowhead_offset=.3,
                 arrowhead_offset_mode='scaled',
                 cylinder_class=None,
                 arrowhead_class=None,
                 **opts):
        if arrowhead_class is None:
            arrowhead_class = self.arrowhead_class
        if cylinder_class is None:
            cylinder_class = self.cylinder_class

        ends = np.asanyarray(ends)
        starts = np.asanyarray(starts)
        arrow_vectors = ends - starts
        norms = nput.vec_norms(arrow_vectors)
        if arrowhead_offset_mode == 'scaled':
            arrowhead_offset = arrowhead_offset * norms
        disp_vectors = arrowhead_offset * nput.vec_normalize(arrow_vectors, norms=norms)
        arrow_starts = ends - disp_vectors
        if arrowhead_radius_mode == 'scaled':
            arrowhead_radius = arrowhead_radius * radius
        arrowheads = arrowhead_class(arrow_starts, ends,
                                     radius=arrowhead_radius,
                                     top_radius=top_radius,
                                     **opts)
        cylinders = cylinder_class(starts, arrow_starts, radius=radius, **opts)

        super().__init__(
            arrowheads, cylinders
        )

class X3DCappedCylinder(X3DGroup):

    cap_class = X3DSphere
    cylinder_class = X3DCylinder

    def __init__(self,
                 starts, ends,
                 radius=1,
                 cylinder_class=None,
                 cap_class=None,
                 cap_offset=0,
                 use_caps=(True, True),
                 **opts):
        if cap_class is None:
            cap_class = self.cap_class
        if cylinder_class is None:
            cylinder_class = self.cylinder_class

        if cap_offset != 0:
            raise NotImplementedError("cap offseting not supported yet")

        if use_caps is True: use_caps = [True, True]
        elif use_caps is False: use_caps = [False, False]

        cap_list = []
        if use_caps[0]:
            cap_list.append(cap_class(starts, radius=radius, **opts))
        if use_caps[1]:
            cap_list.append(cap_class(ends, radius=radius, **opts))
        cylinders = cylinder_class(starts, ends, radius=radius, **opts)

        super().__init__(
            *cap_list, cylinders
        )

class X3DText(X3DGeometryGroup):
    tag_class = X3DHTML.Text

class X3DTorus(X3DGeometryGroup):
    tag_class = X3DHTML.Torus

    def prep_geometry_opts(self, centers, radius=1, inner_radius=None,
                           normal=None, rotation=None, scale=None,
                           angle=None,
                           **opts):
        centers = self.prep_vecs(centers)
        normal = self.prep_vecs(normal, centers.shape[0])
        if angle is not None and angle < 0:
            if rotation is None:
                rotation = [0, 0, 1, 0]
            if isinstance(rotation, str):
                rotation = np.array(rotation.split()).astype(float)
            rotation = list(rotation[:3]) + [rotation[3] + angle]
            angle = abs(angle)
        rotation = self.prep_vecs(rotation, centers.shape[0])
        scale = self.prep_vecs(scale, centers.shape[0])
        radius = self.prep_const(radius, centers.shape[0])
        inner_radius = self.prep_const(inner_radius, centers.shape[0])
        angle = self.prep_const(angle, centers.shape[0])

        return [
            {
                "translation": s, "normal": n, "outerRadius": r,
                "innerRadius": i, 'rotation': rot, 'scale': sc,
                'angle': ang,
                **opts}
            for s, n, rot, sc, r, i, ang in zip(
                centers, normal,
                rotation, scale,
                radius, inner_radius, angle
            )
        ]

class X3DCoordinatesWrapper(X3DGeometryGroup):
    tag_class: X3DHTML.X3DElement
    def create_tag_object(self, *, point, color=None, **etc):
        body = [X3DCoordinate(point).to_x3d()]
        if color is not None:
            body.append(X3DColor(color).to_x3d())
        return self.tag_class(body, **etc)
    def prep_geometry_opts(self, point, **etc):
        return [
            dict({"translation":"0,0,0", "point":point}, **etc)
        ]

class X3DIndexedCoordinatesWrapper(X3DCoordinatesWrapper):
    def prep_geometry_opts(self, point, indices, vertex_colors=None, **etc):
        base_dict = dict(
                {
                    'index': " ".join(np.asanyarray(indices).flatten().astype(int).astype(str))
                },
                **super().prep_geometry_opts(point, **etc)[0]
            )
        if vertex_colors is not None:
            base_dict['colorPerVertex'] = True
            base_dict['color'] = vertex_colors
        return [base_dict]

class X3DGeometry2DGroup(X3DGeometryGroup):
    def prep_geometry_opts(self, center, **etc):
        # raise Exception(crosses)
        # a = np.concatenate([crosses, [angs]])
        # a, _ = self.get_rotation(crosses, up_vector=up_vector)
        return [
            {
                "translation":center,
                 **etc
            }
        ]

class X3DRectangle2D(X3DGeometry2DGroup):
    tag_class = X3DHTML.Rectangle2D
class X3DCircle2D(X3DGeometry2DGroup):
    tag_class = X3DHTML.Circle2D
    def prep_geometry_opts(self, centers, radius=1,
                           normal=None, rotation=None, scale=None,
                           angle=None,
                           **opts):
        centers = self.prep_vecs(centers)
        normal = self.prep_vecs(normal, centers.shape[0])
        if angle is not None and angle < 0:
            if rotation is None:
                rotation = [0, 0, 1, 0]
            if isinstance(rotation, str):
                rotation = np.array(rotation.split()).astype(float)
            rotation = list(rotation[:3]) + [rotation[3] + angle]
            angle = abs(angle)
        rotation = self.prep_vecs(rotation, centers.shape[0])
        scale = self.prep_vecs(scale, centers.shape[0])
        radius = self.prep_const(radius, centers.shape[0])
        angle = self.prep_const(angle, centers.shape[0])

        return [
            {
                "translation": s, "normal": n, "radius": r,
                'rotation': rot, 'scale': sc,
                'angle': ang,
                **opts}
            for s, n, rot, sc, r, ang in zip(
                centers, normal,
                rotation, scale,
                radius, angle
            )
        ]

class X3DDisk2D(X3DGeometry2DGroup):
    tag_class = X3DHTML.Disk2D

    def prep_geometry_opts(self, centers, radius=1, inner_radius=None,
                           normal=None, rotation=None, scale=None,
                           angle=None,
                           **opts):
        centers = self.prep_vecs(centers)
        normal = self.prep_vecs(normal, centers.shape[0])
        if angle is not None and angle < 0:
            if rotation is None:
                rotation = [0, 0, 1, 0]
            if isinstance(rotation, str):
                rotation = np.array(rotation.split()).astype(float)
            rotation = list(rotation[:3]) + [rotation[3] + angle]
            angle = abs(angle)
        rotation = self.prep_vecs(rotation, centers.shape[0])
        scale = self.prep_vecs(scale, centers.shape[0])
        radius = self.prep_const(radius, centers.shape[0])
        inner_radius = self.prep_const(inner_radius, centers.shape[0])
        angle = self.prep_const(angle, centers.shape[0])

        return [
            {
                "translation": s, "normal": n, "outerRadius": r,
                "innerRadius": i, 'rotation': rot, 'scale': sc,
                'angle': ang,
                **opts}
            for s, n, rot, sc, r, i, ang in zip(
                centers, normal,
                rotation, scale,
                radius, inner_radius, angle
            )
        ]
class X3DPolyline2D(X3DGeometry2DGroup):
    tag_class = X3DHTML.Polyline2D

class X3DPointSet(X3DCoordinatesWrapper):
    tag_class = X3DHTML.PointSet
class X3DLine(X3DCoordinatesWrapper):
    tag_class = X3DHTML.LineSet
class X3DTriangleSet(X3DCoordinatesWrapper):
    tag_class = X3DHTML.TriangleSet
class X3DIndexedTriangleSet(X3DIndexedCoordinatesWrapper):
    tag_class = X3DHTML.IndexedTriangleSet
class X3DIndexedLineSet(X3DIndexedCoordinatesWrapper):
    tag_class = X3DHTML.IndexedLineSet
    def prep_geometry_opts(self, point, indices, **etc):
        opts = super().prep_geometry_opts(point, indices, **etc)
        opts[0]['coordIndex'] = opts[0].pop('index')
        return opts
class X3DIndexedQuadSet(X3DIndexedCoordinatesWrapper):
    tag_class = X3DHTML.IndexedQuadSet
class X3DIndexedFaceSet(X3DIndexedCoordinatesWrapper):
    tag_class = X3DHTML.IndexedFaceSet

class X3DListAnimator(X3DGroup):
    def __init__(self, *frames, id=None, animation_duration=2, running=True, slider=False,
                 # allow_interpolation=True,
                 **opts):
        self.uuid = str(uuid.uuid4())
        if id is None:
            id = f"animation-{self.uuid}"
        self.id = id
        anim_frames = X3DSwitch(
                *frames,
                id=id,
                whichChoice="0"
            )
        nframes = len(anim_frames.children) - 1
        key_frames = np.linspace(0, 1, nframes)
        elements = []
        if slider:
            elements.append(
                JHTML.Input(type="range", value="0", min="0", max=f"{nframes}", step="1", cls="slider",
                            oninput=f"""document.getElementById("{id}").setAttribute("whichChoice", this.value)""")
            )
        elements.append(anim_frames)
        if running:
            elements.extend([
                X3DHTML.TimeSensor(id=f'animation-clock-{self.uuid}', cycleInterval=animation_duration, loop=True,
                                   enabled=running
                                   ),
                X3DHTML.IntegerSequencer(id=f'animation-indexer-{self.uuid}', key=key_frames,
                                         keyValue=np.arange(nframes)),
                X3DHTML.Route(
                    fromField='fraction_changed', fromNode=f'animation-clock-{self.uuid}',
                    toField='set_fraction', toNode=f'animation-indexer-{self.uuid}'
                ),
                X3DHTML.Route(
                    fromField='value_changed', fromNode=f'animation-indexer-{self.uuid}',
                    toField='whichChoice', toNode=self.id
                )
            ])

        super().__init__(elements, **opts)




"""
Provides interfaces that _support_ objects
and make it easier to build more reliable,
higher-functioning classes
"""

import abc, os

from .Persistence import PersistenceLocation
from .Checkpointing import NumPyCheckpointer

__all__ = ["BaseObjectManager", "FileBackedObjectManager"]

class BaseObjectManager(metaclass=abc.ABCMeta):
    """
    Defines the basic parameters of an object interface
    that can handle marshalling the core data behind
    and object attribute to disk or vice versa
    """

    def __init__(self, obj):
        self.obj = obj
        self._base_name = None

    def get_basename(self):
        if hasattr(self.obj, 'serialization_id'):
            obj_id = self.obj.serialization_id
        else:
            obj_id = id(self.obj)
        return "{}_{}".format(type(self.obj).__name__, obj_id)

    @property
    def basename(self):
        if self._base_name is None:
            self._base_name = self.get_basename()
        return self._base_name

    @abc.abstractmethod
    def save_attr(self, attr):
        """
        Saves some attribute of the object

        :param attr:
        :type attr:
        :return:
        :rtype:
        """
        raise NotImplementedError("abstract interface")

    @abc.abstractmethod
    def load_attr(self, attr):
        """
        Loads some attribute of the object

        :param attr:
        :type attr:
        :return:
        :rtype:
        """
        raise NotImplementedError("abstract interface")

    @abc.abstractmethod
    def del_attr(self, attr):
        """
        Deletes some attribute of the object

        :param attr:
        :type attr:
        :return:
        :rtype:
        """
        raise NotImplementedError("abstract interface")

class FileBackedObjectManager(BaseObjectManager):
    """
    Provides an interface to back an object with
    a serializer
    """

    default_directory = None
    def __init__(self,
                 obj,
                 chk=None,
                 loc=None,
                 checkpoint_class=NumPyCheckpointer
                 ):
        """
        :param obj: the object to back
        :type obj: object
        :param chk: a checkpointer to manage storing attributes
        :type chk: Checkpointer
        :param loc: the location where attributes should be stored
        :type loc: str
        :param checkpoint_class: a subclass of Checkpointer that implements the actual writing to disk
        :type checkpoint_class: Type[Checkpointer]
        """
        super().__init__(obj)
        if chk is None:
            if loc is None:
                loc = self.get_default_directory().loc
            obj_file = os.path.join(loc, self.basename+checkpoint_class.default_extension)
            chk = checkpoint_class(obj_file)

        self.chk = chk
        self._id = None
        self._cache = {}

    @classmethod
    def get_default_directory(cls):
        if cls.default_directory is None:
            cls.default_directory = PersistenceLocation("file_backed_objects")
        return cls.default_directory

    @property
    def basename(self):
        if self._tag is None:
            self._tag = self.get_basename()
        return self.basename
    @basename.setter
    def basename(self, v):
        self._tag = v

    def get_basename(self):
        if hasattr(self.obj, 'serialization_id'):
            obj_id = self.obj.serialization_id
        else:
            obj_id = id(self.obj)
        return "{}_{}".format(type(self.obj).__name__, obj_id)

    def save_attr(self, attr):
        with self.chk:
            self.chk[attr] = getattr(self.obj, attr)
        return FileBackedAttribute(self, attr)

    def load_attr(self, attr):
        with self.chk:
            return self.chk[attr]

class FileBackedAttribute:
    """
    A helper class to make it very clear that
    an attribute is backed by a file on disk
    """

    def __init__(self, manager, attr):
        self.manager = manager
        self.attr = attr


"""
Provides helpful link to external APIs implemented in JavaScript
"""

__all__ = []
from .d3 import *; from .d3 import __all__ as exposed
__all__ += exposed
from .d3_backend import *; from .d3_backend import __all__ as exposed
__all__ += exposed


import os, enum, weakref

__all__ = [
    "Logger",
    "NullLogger",
    "LogLevel",
    "LoggingBlock"
]


class LogLevel(enum.Enum):
    """
    A simple log level object to standardize more pieces of the logger interface
    """
    Quiet = 0
    Warnings = 1
    Normal = 10
    Debug = 50
    MoreDebug = 75
    All = 100
    Never = 1000 # for debug statements that should really be deleted but I'm too lazy to

    def __eq__(self, other):
        if isinstance(other, LogLevel):
            other = other.value
        return self.value == other
    def __le__(self, other):
        if isinstance(other, LogLevel):
            other = other.value
        return self.value <= other
    def __ge__(self, other):
        if isinstance(other, LogLevel):
            other = other.value
        return self.value >= other
    def __lt__(self, other):
        if isinstance(other, LogLevel):
            other = other.value
        return self.value < other
    def __gt__(self, other):
        if isinstance(other, LogLevel):
            other = other.value
        return self.value > other

class LoggingBlock:
    """
    A class that extends the utility of a logger by automatically setting up a
    named block of logs that add context and can be
    that
    """
    block_settings = [
        {
            'opener': ">>" + "-" * 25 + ' {tag} ' + "-" * 25,
            'prompt': "::{meta} ",
            'closer': '>>'+'-'*50+'<<'
        },
        {
            'opener': "::> {tag}",
            'prompt': "  >{meta} ",
            'closer': '<::'
        }
    ]
    block_level_padding= " " * 2
    def __init__(self,
                 logger,
                 log_level=None,
                 block_level=0,
                 block_level_padding=None,
                 tag=None,
                 opener=None,
                 prompt=None,
                 closer=None,
                 printoptions=None,
                 **tag_vars
                 ):
        self.logger = logger
        if block_level_padding is None:
            block_level_padding = self.block_level_padding
        if block_level >= len(self.block_settings):
            padding = block_level_padding * (block_level - len(self.block_settings) + 1)
            settings = {k: padding + v for k,v in self.block_settings[-1].items()}
        else:
            settings = self.block_settings[block_level]

        self._tag = tag
        self._og_tagvars = tag_vars
        self._tagvars = tag_vars

        self._old_loglev = None
        self.log_level = log_level if log_level is not None else logger.verbosity
        self.opener = settings['opener'] if opener is None else opener
        self._old_prompt = None
        self.prompt = settings['prompt'] if prompt is None else prompt
        self.closer = settings['closer'] if closer is None else closer
        self._in_block = False

        self._print_manager=None
        self.printopts = printoptions

    @property
    def tag(self):
        if self._tag is None:
            self._tag = ""
        elif not isinstance(self._tag, str):
            if callable(self._tag):
                self._tag = self._tag()
            else:
                tag_vars = self._tag[1] # type:dict
                tag_vars = tag_vars.get('preformatter', lambda **kw:kw)(**tag_vars)
                self._tag = self._tag[0].format(**tag_vars)
        elif self._tagvars is not None:
            tag_vars = self._tagvars
            tag_vars = tag_vars.get('preformatter', lambda **kw:kw)(**tag_vars)
            self._tag = self._tag.format(**tag_vars)
            self._tagvars = None

        return self._tag

    def __enter__(self):
        if self.log_level <= self.logger.verbosity:
            self._in_block = True
            self.logger.log_print(self.opener, tag=self.tag, padding="")
            self._old_prompt = self.logger.padding
            self.logger.padding = self.prompt
            self._old_loglev = self.logger.verbosity
            self.logger.verbosity = self.log_level
            self.logger.block_level += 1

            if self.printopts is not None and self._print_manager is None:
                from numpy import printoptions
                self._print_manager = printoptions(**self.printopts)
                self._print_manager.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._in_block:
            self._in_block = False
            self.logger.log_print(self.closer, tag=self.tag, padding="")
            self.logger.padding = self._old_prompt
            self._old_prompt = None
            self.logger.verbosity = self._old_loglev
            self._old_loglev = None
            self.logger.block_level -= 1

            if self._print_manager is not None:
                self._print_manager.__exit__(exc_type, exc_val, exc_tb)
                self._print_manager = None

class Logger:
    """
    Defines a simple logger object to write log data to a file based on log levels.
    """

    LogLevel = LogLevel

    _loggers = weakref.WeakValueDictionary()
    default_verbosity = LogLevel.Normal
    def __init__(self,
                 log_file=None,
                 log_level=None,
                 print_function=None,
                 padding="",
                 newline="\n",
                 repad_messages=True
                 ):
        self.log_file = log_file
        self.verbosity = log_level if log_level is not None else self.default_verbosity
        self.repad = repad_messages
        self.padding = padding
        self.newline = newline
        self.block_level = 0 # just an int to pass to `block(...)` so that it can
        self.auto_flush = True
        if print_function is None:
            print_function = print
        self.print_function = print_function
        self.active = True

    def to_state(self, serializer=None):
        return {
            'log_file': self.log_file,
            'verbosity': self.verbosity,
            'padding': self.padding,
            'newline': self.newline,
            'print_function': None if self.print_function is print else self.print_function
        }
    @classmethod
    def from_state(cls, state, serializer=None):
        return cls(**state)

    def block(self, **kwargs):
        return LoggingBlock(self, block_level=self.block_level, **kwargs)

    def register(self, key):
        """
        Registers the logger under the given key
        :param key:
        :type key:
        :return:
        :rtype:
        """
        self._loggers[key] = self
    @classmethod
    def lookup(cls, key, construct=False):
        """
        Looks up a logger. Has the convenient, but potentially surprising
        behavior that if no logger is found a `NullLogger` is returned.
        :param key:
        :type key:
        :return:
        :rtype:
        """
        if key in cls._loggers:
            logger = cls._loggers[key]
        elif isinstance(key, Logger):
            logger = key
        elif key is True:
            logger = cls()
        else:
            if isinstance(key, str):
                try:
                    ll = LogLevel[key]
                except KeyError:
                    if construct:
                        logger = cls(key)
                    else:
                        logger = None
                else:
                    logger = cls(log_level=ll)
            else:
                logger = None
        if logger is None:
            logger = NullLogger()

        return logger

    @staticmethod
    def preformat_keys(key_functions):
        """
        Generates a closure that will take the supplied
        keys/function pairs and update them appropriately

        :param key_functions:
        :type key_functions:
        :return:
        :rtype:
        """

        def preformat(*args, **kwargs):

            for k,v in kwargs.items():
                if k in key_functions:
                    kwargs[k] = key_functions[k](v)

            return args, kwargs
        return preformat

    def format_message(self, message, *params, preformatter=None, _repad=None, _newline=None, _padding=None, **kwargs):
        if preformatter is not None:
            args = preformatter(*params, **kwargs)
            if isinstance(args, dict):
                kwargs = args
                params = ()
            elif (
                    isinstance(args, tuple)
                    and len(args) == 2
                    and isinstance(args[1], dict)
            ):
                params, kwargs = args
            else:
                params = ()
                kwargs = args

        if _repad is None: _repad = self.repad
        if _repad:
            kwargs = {
                k: (
                    self.pad_newlines(v, newline=_newline, padding=_padding, **kwargs)
                        if isinstance(v, str) else
                    v
                )
                for k,v in kwargs.items()
            }

        if len(kwargs) > 0:
            message = message.format(*params, **kwargs)
        elif len(params) > 0:
            message = message.format(*params)
        return message

    def format_metainfo(self, metainfo):
        if metainfo is None:
            return ""
        else:
            import json
            return json.dumps(metainfo)

    def pad_newlines(self, obj, padding=None, newline=None, **kwargs):
        if not isinstance(obj, str): obj = '{}'.format(obj)
        if padding is None: padding = self.padding
        if newline is None: newline = self.newline
        rep = (newline + padding).format(**kwargs)
        return obj.replace("\n", rep)
    @staticmethod
    def split_lines(obj):
        return str(obj).splitlines()
    @staticmethod
    def prep_array(obj):
        import numpy as np
        with np.printoptions(linewidth=1e8, edgeitems=1e3, threshold=1e8):
            return str(obj).splitlines()
    @staticmethod
    def prep_dict(obj):
        return ["{k}: {v}".format(k=k, v=v) for k,v in obj.items()]

    def log_print(self,
                  message,
                  *messrest,
                  message_prepper=None,
                  padding=None, newline=None,
                  log_level=None,
                  metainfo=None, print_function=None,
                  print_options=None,
                  sep=None, end=None, file=None, flush=None,
                  preformatter=None,
                  **kwargs
                  ):
        """
        :param message: message to print
        :type message: str | Iterable[str]
        :param params:
        :type params:
        :param print_options: options to be passed through to print
        :type print_options:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """


        if log_level is None:
            log_level = self.default_verbosity

        if log_level <= self.verbosity:

            if padding is None:
                padding = self.padding
            if newline is None:
                newline = self.newline
            if print_function is None:
                print_function = self.print_function

            if message_prepper is not None:
                message = message_prepper(message, *messrest)
                messrest = ()

            if len(messrest) > 0:
                message = [message, *messrest]

            if not isinstance(message, str):
                joiner = (newline + padding)
                message = joiner.join(
                    [padding + message[0]]
                    + list(message[1:])
                )
            else:
                message = padding + message

            # print(">>>>", repr(message), params)

            if print_options is None:
                print_options = {}
            if sep is not None:
                print_options['sep'] = sep
            if end is not None:
                print_options['end'] = end
            if file is not None:
                print_options['file'] = file
            if flush is not None:
                print_options['flush'] = flush

            if 'flush' not in print_options:
                print_options['flush'] = self.auto_flush

            if log_level <= self.verbosity:

                msg = self.format_message(message,
                                          meta=self.format_metainfo(metainfo),
                                          preformatter=preformatter,
                                          _newline=newline,
                                          _padding=padding,
                                          **kwargs)
                if isinstance(print_function, str) and print_function == 'echo':
                    if self.log_file is not None:
                        self._print_message(print, msg, self.log_file, print_options)
                    self._print_message(print, msg, None, print_options)
                else:
                    self._print_message(print_function, msg, self.log_file, print_options)
    @staticmethod
    def _print_message(print_function, msg, log, print_options):
        if isinstance(log, str):
            if not os.path.isdir(os.path.dirname(log)):
                try:
                    os.makedirs(os.path.dirname(log))
                except OSError:
                    pass
            # O_NONBLOCK is *nix only
            with open(log, mode="a", buffering=1 if print_options['flush'] else -1) as lf:  # this is potentially quite slow but I am also quite lazy
                print_function(msg, file=lf, **print_options)
        elif log is None:
            print_function(msg, **print_options)
        else:
            print_function(msg, file=log, **print_options)

    def __repr__(self):
        return "{}({}, {})".format(
            type(self).__name__,
            self.log_file,
            self.verbosity
        )

class NullLogger(Logger):
    """
    A logger that implements the interface, but doesn't ever print.
    Allows code to avoid a bunch of "if logger is not None" blocks
    """
    def __init__(self, *log_files, **logger_opts):
        super().__init__(*log_files, **logger_opts)
        self.active = False
    def log_print(self, message, *params, print_options=None, padding=None, newline=None, **kwargs):
        pass
    def __bool__(self):
        return False